@@ -1,9 +1,11 @@
 /*
 **  CWIMAPStore.m
 **
-**  Copyright (c) 2001-2007
+**  Copyright (c) 2001-2007 Ludovic Marcotte
+**  Copyright (C) 2016-2018 Riccardo Mottola
 **
 **  Author: Ludovic Marcotte <ludovic@Sophos.ca>
+**          Riccardo Mottola <rm@gnu.org>
 **
 **  This library is free software; you can redistribute it and/or
 **  modify it under the terms of the GNU Lesser General Public
@@ -15,35 +17,34 @@
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 **  Lesser General Public License for more details.
 **  
-**  You should have received a copy of the GNU Lesser General Public
-**  License along with this library; if not, write to the Free Software
-**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+** You should have received a copy of the GNU General Public License
+** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include <Pantomime/CWIMAPStore.h>
+#import <Pantomime/CWIMAPStore.h>
 
-#include <Pantomime/CWConstants.h>
-#include <Pantomime/CWFlags.h>
-#include <Pantomime/CWFolderInformation.h>
-#include <Pantomime/CWIMAPCacheManager.h>
-#include <Pantomime/CWIMAPFolder.h>
-#include <Pantomime/CWIMAPMessage.h>
-#include <Pantomime/CWMD5.h>
-#include <Pantomime/CWMIMEUtility.h>
-#include <Pantomime/NSData+Extensions.h>
-#include <Pantomime/NSScanner+Extensions.h>
-#include <Pantomime/NSString+Extensions.h>
-#include <Pantomime/CWTCPConnection.h>
-#include <Pantomime/CWURLName.h>
+#import <Pantomime/CWConstants.h>
+#import <Pantomime/CWFlags.h>
+#import <Pantomime/CWFolderInformation.h>
+#import <Pantomime/CWIMAPCacheManager.h>
+#import <Pantomime/CWIMAPFolder.h>
+#import <Pantomime/CWIMAPMessage.h>
+#import <Pantomime/CWMD5.h>
+#import <Pantomime/CWMIMEUtility.h>
+#import <Pantomime/NSData+Extensions.h>
+#import <Pantomime/NSScanner+Extensions.h>
+#import <Pantomime/NSString+Extensions.h>
+#import <Pantomime/CWTCPConnection.h>
+#import <Pantomime/CWURLName.h>
 
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSBundle.h>
-#include <Foundation/NSCharacterSet.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSPathUtilities.h>
-#include <Foundation/NSScanner.h>
-#include <Foundation/NSValue.h>
+#import <Foundation/NSAutoreleasePool.h>
+#import <Foundation/NSBundle.h>
+#import <Foundation/NSCharacterSet.h>
+#import <Foundation/NSException.h>
+#import <Foundation/NSNotification.h>
+#import <Foundation/NSPathUtilities.h>
+#import <Foundation/NSScanner.h>
+#import <Foundation/NSValue.h>
 
 #include <ctype.h>
 #include <stdio.h>
@@ -61,7 +62,7 @@ static NSData *CRLF;
 //
 // "0" means no literal.
 //
-static inline int has_literal(char *buf, int c)
+static inline int has_literal(char *buf, unsigned c)
 {
   char *s;
 
@@ -77,7 +78,7 @@ static inline int has_literal(char *buf, int c)
       d = 1;
       s--;
       
-      while (isdigit(*s))
+      while (isdigit((int)(unsigned char)*s))
 	{
 	  value += ((*s-48) * d);
 	  d *= 10;
@@ -118,22 +119,23 @@ static inline int has_literal(char *buf, int c)
 		  info: (NSDictionary *) theInfo
 {
   self = [super init];
-
-  command = theCommand;
-  literal = 0;
-
-  ASSIGN(arguments, theArguments);
-  ASSIGN(tag, theTag);
-
-  if (theInfo)
+  if (self)
     {
-      info = [[NSMutableDictionary alloc] initWithDictionary: theInfo];
-    }
-  else
-    {
-      info = [[NSMutableDictionary alloc] init];
-    }
+      command = theCommand;
+      literal = 0;
 
+      ASSIGN(arguments, theArguments);
+      ASSIGN(tag, theTag);
+
+      if (theInfo)
+	{
+	  info = [[NSMutableDictionary alloc] initWithDictionary: theInfo];
+	}
+      else
+	{
+	  info = [[NSMutableDictionary alloc] init];
+	}
+    }
   return self;
 }
 
@@ -169,7 +171,7 @@ static inline int has_literal(char *buf, int c)
 - (void) _parseCAPABILITY;
 - (void) _parseEXISTS;
 - (void) _parseEXPUNGE;
-- (void) _parseFETCH: (int) theMSN;
+- (void) _parseFETCH: (unsigned) theMSN;
 - (void) _parseLIST;
 - (void) _parseLSUB;
 - (void) _parseNO;
@@ -206,19 +208,20 @@ static inline int has_literal(char *buf, int c)
   if (thePort == 0) thePort = 143;
 
   self = [super initWithName: theName  port: thePort];
-
-  _folderSeparator = 0;
-  _selectedFolder = nil;
-  _tag = 1;
-
-  _folders = [[NSMutableDictionary alloc] init];
-  _openFolders = [[NSMutableDictionary alloc] init];
-  _subscribedFolders = [[NSMutableArray alloc] init];
-  _folderStatus = [[NSMutableDictionary alloc] init];
-
-  _lastCommand = IMAP_AUTHORIZATION;
-  _currentQueueObject = nil;
-  
+  if (self)
+    {
+      _folderSeparator = 0;
+      _selectedFolder = nil;
+      _tag = 1;
+      
+      _folders = [[NSMutableDictionary alloc] init];
+      _openFolders = [[NSMutableDictionary alloc] init];
+      _subscribedFolders = [[NSMutableArray alloc] init];
+      _folderStatus = [[NSMutableDictionary alloc] init];
+      
+      _lastCommand = IMAP_AUTHORIZATION;
+      _currentQueueObject = nil;
+    }
   return self;
 }
 
@@ -268,7 +271,7 @@ static inline int has_literal(char *buf, int c)
 {
   id aData;
  
-  int i, count;
+  NSUInteger i, count;
   char *buf;
 
   [super updateRead];
@@ -291,9 +294,10 @@ static inline int has_literal(char *buf, int c)
 
 	  if (_currentQueueObject->literal < 0)
 	    {
-	      int x;
+	      NSUInteger x;
 
-	      x = -2-_currentQueueObject->literal;
+              NSAssert(_currentQueueObject->literal <= 2, @"Negative literal too small");
+	      x = (NSUInteger)(-2 - _currentQueueObject->literal);
 	      [[_currentQueueObject->info objectForKey: @"NSData"] appendData: [aData subdataToIndex: x]];
 	      [_responsesFromServer addObject: [aData subdataFromIndex: x]];
 	      //NSLog(@"orig = |%@|, chooped = |%@|   |%@|", [aData asciiString], [[aData subdataToIndex: x] asciiString], [[aData subdataFromIndex: x] asciiString]);
@@ -396,7 +400,8 @@ static inline int has_literal(char *buf, int c)
       //
       if (i == 1)
 	{
-	  int d, j, msn, len;
+	  unsigned d, msn, len;
+          unsigned j;
 	  BOOL b;
 
 	  //
@@ -441,7 +446,7 @@ static inline int has_literal(char *buf, int c)
 	  // Let's see if we can read a MSN
 	  while (j < count && *buf != ' ')
 	    {
-	      if (!isdigit(*buf)) b = NO;
+	      if (!isdigit((int)(unsigned char)*buf)) b = NO;
 	      buf++; j++;
 	    }
 	  
@@ -459,7 +464,7 @@ static inline int has_literal(char *buf, int c)
 	  //
 	  if (b)
 	    {
-	      int k;
+	      unsigned k;
 	      
 	      k = j;
 
@@ -478,7 +483,7 @@ static inline int has_literal(char *buf, int c)
 	      buf += (j-i);
 	      k = j+1;
 
-	      while (k < count && isalpha(*buf))
+	      while (k < count && isalpha((int)(unsigned char)*buf))
 		{
 		  //NSLog(@"response after c = %c", *buf);
 		  buf++; k++;
@@ -611,13 +616,17 @@ static inline int has_literal(char *buf, int c)
 	    {
 	      [self _parseCAPABILITY];
 	    }
+          else
+            {
+              NSDebugLog(@"IMAP Unknown tagged response: %s", buf);
+            }
 	}
       //
       // We got a tagged response
       //
       else
 	{
-	  int j;
+	  unsigned j;
 
 	  //NSData *foo;
 	  //foo = [NSData dataWithBytes: buf-i  length: i];
@@ -703,8 +712,9 @@ static inline int has_literal(char *buf, int c)
     }
   
   // We must verify if we must quote the password
-  if ([thePassword rangeOfCharacterFromSet: [NSCharacterSet punctuationCharacterSet]].length ||
-      [thePassword rangeOfCharacterFromSet: [NSCharacterSet whitespaceCharacterSet]].length)
+  if ( nil != thePassword  &&
+      ([thePassword rangeOfCharacterFromSet: [NSCharacterSet punctuationCharacterSet]].length ||
+      [thePassword rangeOfCharacterFromSet: [NSCharacterSet whitespaceCharacterSet]].length))
     {
       thePassword = [NSString stringWithFormat: @"\"%@\"", thePassword];
     }
@@ -735,7 +745,7 @@ static inline int has_literal(char *buf, int c)
 {
   NSMutableArray *aMutableArray;
   NSString *aString;
-  int i, count;;
+  NSUInteger i, count;;
 
   aMutableArray = [NSMutableArray array];
   count = [_capabilities count];
@@ -926,7 +936,7 @@ static inline int has_literal(char *buf, int c)
 //
 - (NSDictionary *) folderStatus: (NSArray *) theArray
 {
-  int i;
+  NSUInteger i;
 
   [_folderStatus removeAllObjects];
 
@@ -981,7 +991,7 @@ static inline int has_literal(char *buf, int c)
       CWIMAPQueueObject *aQueueObject;
       NSString *aString;
       va_list args;
-      int i, count;
+      NSUInteger i, count;
 
       //NSLog(@"sendCommand invoked, cmd = %i", theCommand);
       va_start(args, theFormat);
@@ -1037,7 +1047,7 @@ static inline int has_literal(char *buf, int c)
   [self writeData: CRLF];
 
   POST_NOTIFICATION(@"PantomimeCommandSent", self, _currentQueueObject->info);
-  PERFORM_SELECTOR_2(_delegate, @selector(commandSent:), @"PantomimeCommandSent", [NSNumber numberWithInt: _lastCommand], @"Command");
+  PERFORM_SELECTOR_2(_delegate, @selector(commandSent:), @"PantomimeCommandSent", [NSNumber numberWithUnsignedInt: _lastCommand], @"Command");
 }
 
 //
@@ -1112,7 +1122,7 @@ static inline int has_literal(char *buf, int c)
 //
 //
 //
-- (unsigned char) folderSeparator
+- (unichar) folderSeparator
 {
   return _folderSeparator;
 }
@@ -1304,7 +1314,7 @@ static inline int has_literal(char *buf, int c)
 
   if (aRange.length)
     {
-      int mark;
+      NSUInteger mark;
 
       mark = aRange.location + 1;
       
@@ -1507,7 +1517,7 @@ static inline int has_literal(char *buf, int c)
   while (![aScanner isAtEnd])
     {
       [aScanner scanUnsignedInt: &value];
-      [aMutableArray addObject: [NSNumber numberWithInt: value]];
+      [aMutableArray addObject: [NSNumber numberWithUnsignedInt: value]];
     }
 
   RELEASE(aScanner);
@@ -1687,11 +1697,11 @@ static inline int has_literal(char *buf, int c)
 - (void) _parseEXISTS
 {
   NSData *aData;
-  int n;
+  unsigned n;
   
   aData = [_responsesFromServer lastObject];
 
-  sscanf([aData cString], "* %d EXISTS", &n);
+  sscanf([aData cString], "* %u EXISTS", &n);
   
   //NSLog(@"_parseExists: %d", n);
 
@@ -1721,7 +1731,8 @@ static inline int has_literal(char *buf, int c)
 {
   CWIMAPMessage *aMessage;
   NSData *aData;
-  int i, msn;
+  NSUInteger i;
+  unsigned msn;
 
   // It looks like some servers send untagged expunge reponses
   // _after_ the selected folder has been closed.
@@ -1732,7 +1743,7 @@ static inline int has_literal(char *buf, int c)
 
   aData = [_responsesFromServer lastObject];
 
-  sscanf([aData cString], "* %d EXPUNGE", &msn);
+  sscanf([aData cString], "* %u EXPUNGE", &msn);
 
   //
   // Messages CAN be expunged before we really had time to FETCH them.
@@ -1867,7 +1878,7 @@ static inline int has_literal(char *buf, int c)
 //
 // And we MUST parse the UID correctly.
 //
-- (void) _parseFETCH: (int) theMSN
+- (void) _parseFETCH: (unsigned) theMSN
 {
   NSMutableString *aMutableString;
   NSCharacterSet *aCharacterSet;
@@ -1879,7 +1890,7 @@ static inline int has_literal(char *buf, int c)
   NSRange aRange;
 
   BOOL done, seen_fetch, must_flush_record;
-  int i, j, count, len;
+  NSUInteger i, j, count, len;
   cache_record r;
   
   //
@@ -1913,7 +1924,7 @@ static inline int has_literal(char *buf, int c)
     {
       aString = [[_responsesFromServer objectAtIndex: i] asciiString];
       //NSLog(@"%i: %@", i, aString);
-      if (!seen_fetch && [aString hasCaseInsensitivePrefix: [NSString stringWithFormat: @"* %d FETCH", theMSN]])
+      if (!seen_fetch && [aString hasCaseInsensitivePrefix: [NSString stringWithFormat: @"* %u FETCH", theMSN]])
 	{
 	  seen_fetch = YES;
 	}
@@ -1934,7 +1945,7 @@ static inline int has_literal(char *buf, int c)
   aCharacterSet = [NSCharacterSet whitespaceAndNewlineCharacterSet];
   len = [aMutableString length];
   aMessage = nil;
-  i = j = 0;
+  i = 0;
   
   aScanner = [[NSScanner alloc] initWithString: aMutableString];
   [aScanner setScanLocation: i];
@@ -1961,9 +1972,9 @@ static inline int has_literal(char *buf, int c)
       //
       if ([aWord characterAtIndex: 0] == '*')
 	{
-	  int msn;
+	  unsigned msn;
 
-	  [aScanner scanInt: &msn];
+	  [aScanner scanUnsignedInt: &msn];
 	  //NSLog(@"msn = %d", msn);
 
 	  //
@@ -2053,10 +2064,10 @@ static inline int has_literal(char *buf, int c)
       //
       else if ([aWord caseInsensitiveCompare: @"RFC822.SIZE"] == NSOrderedSame)
 	{
-	  int size;
+	  unsigned size;
 
-	  [aScanner scanInt: &size];
-	  //NSLog(@"size = %d", size);
+	  [aScanner scanUnsignedInt: &size];
+	  //NSLog(@"size = %u", size);
 	  [aMessage setSize: size];
 	  r.size = size;
 
@@ -2191,7 +2202,8 @@ static inline int has_literal(char *buf, int c)
 {
   NSString *aFolderName, *aString, *theString;
   NSRange r1, r2;
-  int flags, len;
+  int flags;
+  NSUInteger len;
 
   theString = [[_responsesFromServer lastObject] asciiString];
 
@@ -2290,7 +2302,7 @@ static inline int has_literal(char *buf, int c)
 - (void) _parseLSUB
 {
   NSString *aString, *aFolderName;
-  int len;
+  NSUInteger len;
 
   aString = [[NSString alloc] initWithData: [_responsesFromServer lastObject]  encoding: defaultCStringEncoding];
   
@@ -2412,7 +2424,7 @@ static inline int has_literal(char *buf, int c)
     {
       //NSLog(@"REMOVING QUEUE OBJECT");
       
-      [_currentQueueObject->info setObject: [NSNumber numberWithInt: _lastCommand]  forKey: @"Command"];
+      [_currentQueueObject->info setObject: [NSNumber numberWithUnsignedInt: _lastCommand]  forKey: @"Command"];
       POST_NOTIFICATION(@"PantomimeCommandCompleted", self, _currentQueueObject->info);
       PERFORM_SELECTOR_3(_delegate, @selector(commandCompleted:), @"PantomimeCommandCompleted", _currentQueueObject->info);
       
@@ -2622,7 +2634,7 @@ static inline int has_literal(char *buf, int c)
 	// Once the STORE has completed, we update the messages.
 	NSArray *theMessages;
 	CWFlags *theFlags;
-	int i, count;
+	NSUInteger i, count;
 	
 	theMessages = [_currentQueueObject->info objectForKey: @"Messages"];
 	theFlags = [_currentQueueObject->info objectForKey: @"Flags"];
@@ -2656,7 +2668,7 @@ static inline int has_literal(char *buf, int c)
   if (![aData hasCPrefix: "*"])// || _lastCommand == IMAP_AUTHORIZATION)
     {
       //NSLog(@"REMOVING QUEUE OBJECT");      
-      [_currentQueueObject->info setObject: [NSNumber numberWithInt: _lastCommand]  forKey: @"Command"];
+      [_currentQueueObject->info setObject: [NSNumber numberWithUnsignedInt: _lastCommand]  forKey: @"Command"];
       POST_NOTIFICATION(@"PantomimeCommandCompleted", self, _currentQueueObject->info);
       PERFORM_SELECTOR_3(_delegate, @selector(commandCompleted:), @"PantomimeCommandCompleted", _currentQueueObject->info);
       
@@ -2731,7 +2743,7 @@ static inline int has_literal(char *buf, int c)
 {
   CWIMAPMessage *aMessage;
   NSArray *allResults;
-  int i, count;
+  NSUInteger i, count;
   BOOL b;
 
   allResults = [self _uniqueIdentifiersFromData: [_responsesFromServer objectAtIndex: 0]];
@@ -2764,9 +2776,9 @@ static inline int has_literal(char *buf, int c)
       //NSLog(@"Folder count (to remove UID) = %d", [_selectedFolder->allMessages count]);
       b = NO;
 
-      for (i = ([_selectedFolder->allMessages count]-1); i >= 0; i--)
+      for (i = ([_selectedFolder->allMessages count]); i > 0; i--)
 	{   
-	  aMessage = [_selectedFolder->allMessages objectAtIndex: i];
+	  aMessage = [_selectedFolder->allMessages objectAtIndex: i-1];
 	  //aMessage = [theCache objectAtIndex: i];
       
 	  if ([aMessage folder] == nil)
@@ -2846,7 +2858,7 @@ static inline int has_literal(char *buf, int c)
 - (void) _parseSELECT
 {
   NSData *aData;
-  int i, count;
+  NSUInteger i, count;
 
   // The last object in _responsesFromServer is a tagged OK response.
   // We need to parse it here.
@@ -2914,14 +2926,14 @@ static inline int has_literal(char *buf, int c)
   NSData *aData;
 
   NSRange aRange;
-  int messages, unseen;
+  unsigned messages, unseen;
 
   aData = [_responsesFromServer lastObject];
   
   aRange = [aData rangeOfCString: "("  options: NSBackwardsSearch];
   aFolderName = [[[aData subdataToIndex: (aRange.location-1)] subdataFromIndex: 9] asciiString];
   
-  sscanf([[aData subdataFromIndex: aRange.location] cString], "(MESSAGES %d UNSEEN %d)", &messages, &unseen);
+  sscanf([[aData subdataFromIndex: aRange.location] cString], "(MESSAGES %u UNSEEN %u)", &messages, &unseen);
   
   aFolderInformation = [[CWFolderInformation alloc] init];
   [aFolderInformation setNbOfMessages: messages];
