@@ -1,9 +1,13 @@
 /*
 **  CWLocalCacheManager.m
 **
-**  Copyright (c) 2001-2007
+**  Copyright (c) 2001-2007 Ludovic Marcotte
+**  Copyright (C) 2013-2018 Riccardo Mottola
 **
 **  Author: Ludovic Marcotte <ludovic@Sophos.ca>
+**          Riccardo Mottola <rm@gnu.org>
+**          Sebastian Reitenbach
+**
 **
 **  This library is free software; you can redistribute it and/or
 **  modify it under the terms of the GNU Lesser General Public
@@ -15,32 +19,32 @@
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 **  Lesser General Public License for more details.
 **  
-**  You should have received a copy of the GNU Lesser General Public
-**  License along with this library; if not, write to the Free Software
-**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+** You should have received a copy of the GNU General Public License
+** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include <Pantomime/CWLocalCacheManager.h>
+#import <Pantomime/CWLocalCacheManager.h>
 
-#include <Pantomime/io.h>
-#include <Pantomime/CWConstants.h>
-#include <Pantomime/CWFlags.h>
-#include <Pantomime/CWLocalFolder.h>
-#include <Pantomime/CWLocalMessage.h>
-#include <Pantomime/CWParser.h>
-#include <Pantomime/NSData+Extensions.h>
+#import <Pantomime/io.h>
+#import <Pantomime/CWConstants.h>
+#import <Pantomime/CWFlags.h>
+#import <Pantomime/CWLocalFolder.h>
+#import <Pantomime/CWLocalMessage.h>
+#import <Pantomime/CWParser.h>
+#import <Pantomime/NSData+Extensions.h>
 
-#include <Foundation/NSArchiver.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSFileManager.h>
-#include <Foundation/NSNull.h>
-#include <Foundation/NSValue.h>
+#import <Foundation/NSArchiver.h>
+#import <Foundation/NSException.h>
+#import <Foundation/NSFileManager.h>
+#import <Foundation/NSNull.h>
+#import <Foundation/NSValue.h>
 
 #include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>  // For open() and friends.
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <errno.h>
 #include <unistd.h>     // For lseek()
 #include <netinet/in.h> // For ntohl()
 
@@ -78,7 +82,8 @@ static unsigned short version = 1;
   BOOL broken;
 
   self = [super initWithPath: thePath];
-
+  if (self)
+    {
   // We get the attributes of the mailbox
   if ([theFolder type] == PantomimeFormatMbox)
     {
@@ -90,8 +95,8 @@ static unsigned short version = 1;
 						   traverseLink: NO];
     }
 
-  d = [[attributes objectForKey: NSFileModificationDate] timeIntervalSince1970];
-  s = [[attributes objectForKey: NSFileSize] intValue];
+  d = (unsigned int)[[attributes objectForKey: NSFileModificationDate] timeIntervalSince1970];
+  s = (unsigned int)[[attributes objectForKey: NSFileSize] unsignedLongValue];
   broken = NO;
 
   // We get the attribtes of the cache
@@ -115,25 +120,40 @@ static unsigned short version = 1;
   // If the cache exists, lets parse it.
   if ([[attributes objectForKey: NSFileSize] intValue])
     {
-      unsigned int i;
-
       v = read_unsigned_short(_fd);
 
       // HACK: We IGNORE all the previous cache.
       if (v != version)
 	{
 	  //NSLog(@"Ignoring the old cache format.");
-	  ftruncate(_fd, 0);
+	  if (ftruncate(_fd, 0) == -1)
+            {
+
+              if (errno == EACCES || errno == EROFS)
+                NSLog(@"UNABLE TO TRUNCATE CACHE FILE WITH OLD VERSION, NOT WRITABLE");
+              else
+                NSLog(@"UNABLE TO TRUNCATE CACHE FILE WITH OLD VERSION");
+              close(_fd);
+              abort();
+            }
 	  [self synchronize];
 	  return self;
 	}
 
-      _count = read_unsigned_int(_fd);
-      _modification_date = read_unsigned_int(_fd);
+      if (read_unsigned_int(_fd, &_count) <= 0)
+        {
+          NSLog(@"CWLocalCacheManager initWithPath, error reading _count");
+          broken = YES;
+        }
+      read_unsigned_int(_fd, &_modification_date);
 
       if ([(CWLocalFolder *)_folder type] == PantomimeFormatMbox)
 	{
-	  _size = read_unsigned_int(_fd);
+	  if (read_unsigned_int(_fd, &_size) <= 0)
+            {
+              NSLog(@"CWLocalCacheManager initWithPath: failed to read size");
+              broken = YES;
+            }
 	  
 	  if (s != _size || d != _modification_date) broken = YES;
 	}
@@ -148,25 +168,29 @@ static unsigned short version = 1;
  
       if (broken)
 	{
-	  //NSLog(@"Broken cache, we must invalidate.");
+	  NSDebugLog(@"Broken cache, we must invalidate.");
 	  _count = _size = 0;
-	  ftruncate(_fd,0);
+          if (ftruncate(_fd, 0) == -1)
+            {
+
+              if (errno == EACCES || errno == EROFS)
+                NSLog(@"UNABLE TO TRUNCATE BROKEN CACHE FILE, NOT WRITABLE");
+              else
+                NSLog(@"UNABLE TO TRUNCATE BROKEN CACHE FILE");
+              close(_fd);
+              abort();
+            }
 	  [self synchronize];
 	  return self;
 	}
       
       //NSLog(@"Version = %i  date  = %d  size = %d count = %d", v, d, _size, _count);
-
-      for (i = 0; i < _count; i++)
-	{
-	  [((CWFolder *)_folder)->allMessages addObject: AUTORELEASE([[CWLocalMessage alloc] init])];
-	}
     }
   else
     {
       [self synchronize];
     }
-
+    }
   return self;
 }
 
@@ -182,136 +206,119 @@ static unsigned short version = 1;
   [super dealloc];
 }
 
+//
+//
+//
+- (void) readAllMessages
+{
+  [self readMessagesInRange:NSMakeRange(0, NSUIntegerMax)];
+}
 
 //
 // If this method is invoked on already initialized messages, we use
 // the file position or filename field in order to determine if we already
 // initialized or not the message.
 //
-- (void) initInRange: (NSRange) theRange
+- (void) readMessagesInRange: (NSRange) theRange
 {
   CWLocalMessage *aMessage;
-  CWFlags *theFlags;
   
-  unsigned short int len, tot;
-  unsigned char *r, *s;
-  int begin, end, i;
-  BOOL b;
+  unsigned short int tot;
+  unsigned char *r;
+  NSUInteger begin, end, i;
 
   begin = theRange.location;
   end = (NSMaxRange(theRange) <= _count ? NSMaxRange(theRange) : _count);
 
   if (lseek(_fd, ([(CWLocalFolder *)_folder type] == PantomimeFormatMbox) ? 14L : 10L, SEEK_SET) < 0)
     {
-      NSLog(@"lseek failed in initInRange:");
+      NSLog(@"CWLocalCacheManager readMessagesInRange: lseek failed");
       abort();
     }
   
   //NSLog(@"init from %d to %d, count = %d, size of char %d", begin, end, _count, sizeof(char));
 
-  s = (unsigned char *)malloc(65536);
-
   // We MUST skip the last few bytes...
   for (i = begin; i < end ; i++)
     {
-      // aMessage = [[CWLocalMessage alloc] init];
-      aMessage = [((CWFolder *)_folder)->allMessages objectAtIndex: i];
-      [aMessage setFolder: _folder];
-      [aMessage setMessageNumber: i+1];
-      b = NO;
-
-      // With read: 1.157
-      // With read+malloc+free: 1.201
-      // Without read, malloc, free + in-memory parsing: .847
-      // Inline: .892 (not worth it)
-      // With 'static' buffer: .839 (not worth it)
+      uint16_t len;
+      cache_record cr;
+      unsigned int record_length;
+      NSString *mailFilename;
+    
 
       // We parse the record length, date, flags, position in file and the size.
-      len = read_unsigned_int(_fd);
-
-      r = (unsigned char *)malloc(len-4);
+      if (read_unsigned_int(_fd, &record_length) <= 0)
+        {
+          NSLog(@"CWLocalCacheManager readMessagesInRange: failed to read record length (%lu/%lu)", (unsigned long)i, (unsigned long)end);
+          break;
+        }
       
-      if (read(_fd, r, len-4) < 0) { NSLog(@"read failed"); abort(); }
+      r = (unsigned char *)malloc(record_length-4);
+      
+      if (read(_fd, r, record_length-4) < 0) { NSLog(@"read failed"); abort(); }
 
-      theFlags = AUTORELEASE([[CWFlags alloc] initWithFlags: read_unsigned_int_memory(r)]);
-      [aMessage setReceivedDate: [NSCalendarDate dateWithTimeIntervalSince1970: read_unsigned_int_memory(r+4)]];
+      tot = 0;
+      
+      cr.flags = read_unsigned_int_memory(r);
+      tot += 4;
+      
+      cr.date = read_unsigned_int_memory(r+4);
+      tot += 4;
 
+      mailFilename= nil;
+      cr.position = 0;
       if ([(CWLocalFolder *)_folder type] == PantomimeFormatMbox)
 	{
-	  if ([aMessage filePosition] == 0)
-	    {
-	      [aMessage setFilePosition: read_unsigned_int_memory(r+8)];
-	      [aMessage setSize: read_unsigned_int_memory(r+12)];
-	      b = YES;
-	    }
-	  tot = 16;
+	  cr.position = (long)read_unsigned_int_memory(r+tot);
+	  tot += 4;
 	}
       else
 	{
-	  read_string_memory(r+8, s, &len);
-	  if (![aMessage filename])
-	    {
-	      [aMessage setMailFilename: [NSString stringWithUTF8String: (const char *)s]];
-	      [aMessage setSize: read_unsigned_int_memory(r+len+10)];
-	      b = YES;
-	    }
-	  tot = len+14;
+	  mailFilename = read_string_memory(r+tot, &len);
+	  tot += len+2;
 	}
 
-      if (!b)
-	{
-	  free(r);
-	  continue;
-	}
-
-      // We set the flags, only if we need to as they might have
-      // changed since the last time this method was called.
-      [aMessage setFlags: theFlags];
-
-      // .209ms
-      read_string_memory(r+tot, s, &len);
-      [CWParser parseFrom: [NSData dataWithBytes: s  length: len]  inMessage: aMessage  quick: YES];
+      cr.size = read_unsigned_int_memory(r+tot);
+      tot += 4;
+      
+      cr.from = read_data_memory(r+tot, &len);
       tot += len+2;
      
-      // .462ms (+253ms), .217ms
-      read_string_memory(r+tot, s, &len);
-      [CWParser parseInReplyTo: [NSData dataWithBytes: s  length: len]  inMessage: aMessage  quick: YES];
+      cr.in_reply_to = read_data_memory(r+tot, &len);
       tot += len+2;
       
-      // .468ms (+4ms)
-      read_string_memory(r+tot, s, &len);
-      [CWParser parseMessageID: [NSData dataWithBytes: s  length: len]  inMessage: aMessage  quick: YES];
+      cr.message_id = read_data_memory(r+tot, &len);
       tot += len+2;
 
-      // .560ms (+92ms)
-      read_string_memory(r+tot, s, &len);
-      [CWParser parseReferences: [NSData dataWithBytes: s  length: len]  inMessage: aMessage  quick: YES];
+      cr.references = read_data_memory(r+tot, &len);
       tot += len+2;
 
-      // .731ms (+171ms), .135ms
-      read_string_memory(r+tot, s, &len);
-      [CWParser parseSubject:  [NSData dataWithBytes: s  length: len]  inMessage: aMessage  quick: YES];
+      cr.subject = read_data_memory(r+tot, &len);
       tot += len+2;
       
-      // .823ms (+92ms), .80ms
-      read_string_memory(r+tot, s, &len);
-      [CWParser parseDestination: [NSData dataWithBytes: s  length: len]
-		forType: PantomimeToRecipient
-		inMessage: aMessage
-		quick: YES];
+      cr.to = read_data_memory(r+tot, &len);
       tot += len+2;
       
-      // .948ms (+125 to +163ms)
-      read_string_memory(r+tot, s, &len);
-      [CWParser parseDestination: [NSData dataWithBytes: s  length: len]
-		forType: PantomimeCcRecipient
-		inMessage: aMessage
-		quick: YES];
-
+      cr.cc = read_data_memory(r+tot, &len);
+      
+      // zero unused fields
+      cr.imap_uid = 0;
+      cr.filename = NULL;
+      cr.pop3_uid = NULL;
+      
+      aMessage = [[CWLocalMessage alloc] initWithCacheRecord:cr];
+      [aMessage setFolder: _folder];
+ 
+      if ([aMessage filename] != nil)
+	[aMessage setMailFilename: mailFilename];
+  
+      [aMessage setMessageNumber: i+1];
+      [((CWFolder *)_folder)->allMessages addObject: aMessage];
+      RELEASE(aMessage);
+      
       free(r);
     }
-
-  free(s);
 }
 
 //
@@ -319,12 +326,12 @@ static unsigned short version = 1;
 //
 - (NSDate *) modificationDate
 {
-  return [NSDate dateWithTimeIntervalSince1970: _modification_date];
+  return [NSDate dateWithTimeIntervalSince1970:(NSTimeInterval) _modification_date];
 }
 
 - (void) setModificationDate: (NSDate *) theDate
 {
-  _modification_date = [theDate timeIntervalSince1970];
+  _modification_date = (unsigned int)[theDate timeIntervalSince1970];
 }
 
 //
@@ -348,7 +355,7 @@ static unsigned short version = 1;
   NSDictionary *attributes; 
   CWLocalMessage *aMessage;
   unsigned int len, flags;
-  int i;
+  NSUInteger i;
 
   if ([(CWLocalFolder *)_folder type] == PantomimeFormatMbox)
     {
@@ -361,7 +368,7 @@ static unsigned short version = 1;
 						   traverseLink: NO];
     }
   
-  _modification_date = [[attributes objectForKey: NSFileModificationDate] timeIntervalSince1970];
+  _modification_date = (unsigned int)[[attributes objectForKey: NSFileModificationDate] timeIntervalSince1970];
   _count = [_folder->allMessages count];
 
   if (lseek(_fd, 0L, SEEK_SET) < 0)
@@ -371,7 +378,7 @@ static unsigned short version = 1;
   
   // We write our cache version, count, modification date and size.
   write_unsigned_short(_fd, version);
-  write_unsigned_int(_fd, _count);
+  write_unsigned_int(_fd, (uint32_t)_count);
   write_unsigned_int(_fd, _modification_date);
 
   if ([(CWLocalFolder *)_folder type] == PantomimeFormatMbox)
@@ -384,7 +391,11 @@ static unsigned short version = 1;
   //NSLog(@"Synching flags for mailbox %@, count = %d", [(CWLocalFolder *)_folder path], _count);
   for (i = 0; i < _count; i++)
     {
-      len = read_unsigned_int(_fd);
+      if (read_unsigned_int(_fd, &len) <= 0)
+        {
+          NSLog(@"CSWLocalCacheManger synchronize (%lu/%lu)", (unsigned long)i, (unsigned long)_count);
+          break;
+        }
       //NSLog(@"len = %d", len);
 
       if ((NSNull *)(aMessage = [_folder->allMessages objectAtIndex: i]) != [NSNull null])
@@ -430,22 +441,22 @@ static unsigned short version = 1;
   // first five fields, which is 20 bytes long and is added
   // at the very end)
   len = 0;
-  len += [theRecord->from length]+2;
-  len += [theRecord->in_reply_to length]+2;
-  len += [theRecord->message_id length]+2;
-  len += [theRecord->references length]+2;
-  len += [theRecord->subject length]+2;
-  len += [theRecord->to length]+2;
-  len += [theRecord->cc length]+2;
-
-  if ([(CWLocalFolder *)_folder type] == PantomimeFormatMaildir)
+  len += (unsigned int)[theRecord->from length]+2;
+  len += (unsigned int)[theRecord->in_reply_to length]+2;
+  len += (unsigned int)[theRecord->message_id length]+2;
+  len += (unsigned int)[theRecord->references length]+2;
+  len += (unsigned int)[theRecord->subject length]+2;
+  len += (unsigned int)[theRecord->to length]+2;
+  len += (unsigned int)[theRecord->cc length]+2;
+  
+  if ([(CWLocalFolder *)_folder type] == PantomimeFormatMbox)
     {
-      len += strlen(theRecord->filename)+2;
-      len += 16;
+      len += 20;
     }
   else
     {
-      len += 20;
+      len += strlen(theRecord->filename)+2;
+      len += 16;
     }
 
   // We write the length of our entry
@@ -457,23 +468,23 @@ static unsigned short version = 1;
 
   if ([(CWLocalFolder *)_folder type] == PantomimeFormatMbox)
     {
-      write_unsigned_int(_fd, theRecord->position);
+      write_unsigned_int(_fd, (uint32_t)theRecord->position);
     }
   else
     {
-      write_string(_fd, (unsigned char *)theRecord->filename, strlen(theRecord->filename));
+      write_string(_fd, (unsigned char *)theRecord->filename, (uint16_t)strlen(theRecord->filename));
     }
   
   write_unsigned_int(_fd, theRecord->size);
   
   // We write the read of our cached headers (From, In-Reply-To, Message-ID, References, Subject and To)
-  write_string(_fd, (unsigned char *)[theRecord->from bytes], [theRecord->from length]);
-  write_string(_fd, (unsigned char *)[theRecord->in_reply_to bytes], [theRecord->in_reply_to length]);
-  write_string(_fd, (unsigned char *)[theRecord->message_id bytes], [theRecord->message_id length]);
-  write_string(_fd, (unsigned char *)[theRecord->references bytes], [theRecord->references length]);
-  write_string(_fd, (unsigned char *)[theRecord->subject bytes], [theRecord->subject length]);
-  write_string(_fd, (unsigned char *)[theRecord->to bytes], [theRecord->to length]);
-  write_string(_fd, (unsigned char *)[theRecord->cc bytes], [theRecord->cc length]);
+  write_data(_fd, theRecord->from);
+  write_data(_fd, theRecord->in_reply_to);
+  write_data(_fd, theRecord->message_id);
+  write_data(_fd, theRecord->references);
+  write_data(_fd, theRecord->subject);
+  write_data(_fd, theRecord->to);
+  write_data(_fd, theRecord->cc);
 
   _count++;
 }
@@ -491,8 +502,9 @@ static unsigned short version = 1;
 {
   NSDictionary *attributes;
   CWLocalMessage *aMessage;
-
-  unsigned int cache_size, flags, i, len, total_deleted, total_length, type, v;
+  off_t cache_size;
+  NSUInteger i;
+  unsigned int flags, len, total_deleted, total_length, type, v;
   short delta;
   char *buf;
 
@@ -520,7 +532,11 @@ static unsigned short version = 1;
 
   for (i = 0; i < _count; i++)
     {
-      len = read_unsigned_int(_fd);
+      if (read_unsigned_int(_fd, &len) <= 0)
+        {
+          NSLog(@"CWLocalCacheManager expunge: record length read error (%lu/%lu)", (unsigned long)i, (unsigned long)_count);
+          break;
+        }
       aMessage = [_folder->allMessages objectAtIndex: i];
       flags = ((CWFlags *)[aMessage flags])->flags;
       delta = 0;
@@ -595,7 +611,7 @@ static unsigned short version = 1;
 
 	  // We write back our record length, adjusting its size if we need
 	  // to, in the case we are handling a maildir-based cache.
-	  len += delta;
+	  len += (unsigned int)delta;
 	  v = htonl(len);
 	  memcpy((buf+total_length), (char *)&v, 4);
 
@@ -614,15 +630,15 @@ static unsigned short version = 1;
   _count -= total_deleted;
 
   write_unsigned_short(_fd, version);
-  write_unsigned_int(_fd, _count);
+  write_unsigned_int(_fd, (uint32_t)_count);
 
   if ([(CWLocalFolder *)_folder type] == PantomimeFormatMbox)
       {
 	attributes = [[NSFileManager defaultManager] fileAttributesAtPath: [(CWLocalFolder *)_folder path]
 						     traverseLink: NO];
 	
-	_modification_date = [[attributes objectForKey: NSFileModificationDate] timeIntervalSince1970];
-	_size = [[attributes objectForKey: NSFileSize] intValue];
+	_modification_date = (unsigned int)[[attributes fileModificationDate] timeIntervalSince1970];
+	_size = (unsigned int)[attributes fileSize];
 	write_unsigned_int(_fd, _modification_date);
 	write_unsigned_int(_fd, _size);
       }
@@ -630,16 +646,38 @@ static unsigned short version = 1;
     {
       attributes = [[NSFileManager defaultManager] fileAttributesAtPath: [NSString stringWithFormat: @"%@/cur", [(CWLocalFolder *)_folder path]]
 						   traverseLink: NO];
-      _modification_date = [[attributes objectForKey: NSFileModificationDate] timeIntervalSince1970];
+      _modification_date = (unsigned int)[[attributes fileModificationDate] timeIntervalSince1970];
       _size = 0;
       write_unsigned_int(_fd, _modification_date);
     }
   
   // We write our memory cache
-  write(_fd, buf, total_length);
+    if (write(_fd, buf, total_length) != (ssize_t)total_length)
+    {
+      if (errno == EAGAIN)
+        {
+          // Perhaps we could handle this more gracefully?
+          NSLog(@"EXPUNGE CACHE: WRITE OUT ERROR, EAGAIN");
+        }
+      else
+        {
+          NSLog(@"EXPUNGE CACHE: WRITE OUT INCOMPLETE");
+        }
+      abort();
+    }
 
   //ftruncate(_fd, _size);
-  ftruncate(_fd, cache_size);
+  if (ftruncate(_fd, cache_size) == -1)
+    {
+      
+      if (errno == EACCES || errno == EROFS)
+        NSLog(@"UNABLE TO EXPUNGE CACHE, NOT WRITABLE");
+      else if (errno == EFBIG)
+        NSLog(@"UNABLE TO EXPUNGE CACHE, EFBIG");
+      else
+        NSLog(@"UNABLE TO EXPUNGE CACHE");
+      abort();
+    }
   free(buf);
 
   //NSLog(@"Done!");
