@@ -1,9 +1,11 @@
 /*
 **  CWLocalFolder.m
 **
-**  Copyright (c) 2001-2007
+**  Copyright (c) 2001-2007 Ludovic Marcotte
+**  Copyright (C) 2017      Riccardo Mottola
 **
 **  Author: Ludovic Marcotte <ludovic@Sophos.ca>
+**          Riccardo Mottola <rm@gnu.org>
 **
 **  This library is free software; you can redistribute it and/or
 **  modify it under the terms of the GNU Lesser General Public
@@ -15,31 +17,31 @@
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 **  Lesser General Public License for more details.
 **  
-**  You should have received a copy of the GNU Lesser General Public
-**  License along with this library; if not, write to the Free Software
-**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+** You should have received a copy of the GNU General Public License
+** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
-#include <Pantomime/CWLocalFolder.h>
 
-#include <Pantomime/CWConstants.h>
-#include <Pantomime/CWRegEx.h>
-#include <Pantomime/CWFlags.h>
-#include <Pantomime/CWInternetAddress.h>
-#include <Pantomime/CWLocalCacheManager.h>
-#include <Pantomime/CWLocalFolder+maildir.h>
-#include <Pantomime/CWLocalFolder+mbox.h>
-#include <Pantomime/CWLocalMessage.h>
-#include <Pantomime/CWLocalStore.h>
-#include <Pantomime/CWMIMEMultipart.h>
-#include <Pantomime/NSData+Extensions.h>
-#include <Pantomime/NSFileManager+Extensions.h>
-#include <Pantomime/NSString+Extensions.h>
+#import <Pantomime/CWLocalFolder.h>
 
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSHost.h>
-#include <Foundation/NSNotification.h>
-#include <Foundation/NSPathUtilities.h>
+#import <Pantomime/CWConstants.h>
+#import <Pantomime/CWRegEx.h>
+#import <Pantomime/CWFlags.h>
+#import <Pantomime/CWInternetAddress.h>
+#import <Pantomime/CWLocalCacheManager.h>
+#import <Pantomime/CWLocalFolder+maildir.h>
+#import <Pantomime/CWLocalFolder+mbox.h>
+#import <Pantomime/CWLocalMessage.h>
+#import <Pantomime/CWLocalStore.h>
+#import <Pantomime/CWMIMEMultipart.h>
+#import <Pantomime/NSData+Extensions.h>
+#import <Pantomime/NSFileManager+Extensions.h>
+#import <Pantomime/NSString+Extensions.h>
+
+#import <Foundation/NSAutoreleasePool.h>
+#import <Foundation/NSException.h>
+#import <Foundation/NSHost.h>
+#import <Foundation/NSNotification.h>
+#import <Foundation/NSPathUtilities.h>
 
 #include <ctype.h>
 #include <fcntl.h>
@@ -49,13 +51,6 @@
 #include <unistd.h>   // FIXME remove
 #import <stdlib.h> // for rand()
 
-#define FOLDER_APPEND_FAILED \
-  aDictionary = (theFlags ? [NSDictionary dictionaryWithObjectsAndKeys: theData, @"NSData", self, @"Folder", theFlags, @"Flags", nil] : \
-		 [NSDictionary dictionaryWithObjectsAndKeys: theData, @"NSData", self, @"Folder", nil]); \
-	  PERFORM_SELECTOR_3([[self store] delegate], @selector(folderAppendFailed:), PantomimeFolderAppendFailed, aDictionary); \
-  RELEASE(aMutableData); \
-  RELEASE(pool); \
-  return;
 
 //
 // Private methods
@@ -80,35 +75,38 @@
 
   self = [super initWithName: [thePath lastPathComponent]];
 
-  // We initialize those ivars in order to make sure we don't call
-  // the assertion handler when using a maildir-based mailbox.
-  stream = NULL;
-  fd = -1;
+  if (self)
+    {
+      // We initialize those ivars in order to make sure we don't call
+      // the assertion handler when using a maildir-based mailbox.
+      stream = NULL;
+      fd = -1;
 
-  [self setPath: thePath];
+      [self setPath: thePath];
    
-  if ([[NSFileManager defaultManager] fileExistsAtPath: [NSString stringWithFormat: @"%@/new", _path]  isDirectory: &b] && b)
-    {
-      [self setType: PantomimeFormatMaildir];
-    }
-  else
-    {
-      [self setType: PantomimeFormatMbox];
-
-      // We verify if a <name>.tmp was present. If yes, we simply remove it.
-      if ([[NSFileManager defaultManager] fileExistsAtPath: [thePath stringByAppendingString: @".tmp"]])
+      if ([[NSFileManager defaultManager] fileExistsAtPath: [NSString stringWithFormat: @"%@/new", _path]  isDirectory: &b] && b)
 	{
-	  [[NSFileManager defaultManager] removeFileAtPath: [thePath stringByAppendingString: @".tmp"]
-					  handler: nil];
+	  [self setType: PantomimeFormatMaildir];
+	}
+      else
+	{
+	  [self setType: PantomimeFormatMbox];
+
+	  // We verify if a <name>.tmp was present. If yes, we simply remove it.
+	  if ([[NSFileManager defaultManager] fileExistsAtPath: [thePath stringByAppendingString: @".tmp"]])
+	    {
+	      [[NSFileManager defaultManager] removeFileAtPath: [thePath stringByAppendingString: @".tmp"]
+						       handler: nil];
+	    }
+	}
+
+      if ((_type == PantomimeFormatMbox) && ![self open_mbox])
+	{
+	  AUTORELEASE(self);
+	  return nil;
 	}
-    }
 
-  if ((_type == PantomimeFormatMbox) && ![self open_mbox])
-    {
-      AUTORELEASE(self);
-      return nil;
     }
-  
   return self;
 }
 
@@ -333,74 +331,23 @@
   return PantomimeReadWriteMode;
 }
 
-
 //
 // This method is used to append a message to this folder. The message
 // must be specified in raw source. The message is appended to the 
-// local file and is initialized after.
+// stream that must be passed open. The caller has resposibility of
+// eventualy closing it. BOOL returned YES if successful, NO otherwise
 //
-- (void) appendMessageFromRawSource: (NSData *) theData
+- (BOOL) appendMessageFromRawSource: (NSData *) theData
+			   toStream: (FILE *) theStream
+                       withMailFile: (NSString *) mailFile
                               flags: (CWFlags *) theFlags
+
 {
-  NSString *aMailFile, *aMailFilePath;
   NSMutableData *aMutableData;
-  NSDictionary *aDictionary;
   CWLocalMessage *aMessage;
-  NSAutoreleasePool *pool;
-  FILE *aStream;
+  long file_position;
   
-  long mark, file_position;
-
-  pool = [[NSAutoreleasePool alloc] init];
-
   aMutableData = [[NSMutableData alloc] initWithData: theData];
-  aMailFile = nil;
-  aStream = NULL;
-
-  if (!theData || [theData length] == 0)
-    {
-      FOLDER_APPEND_FAILED;
-    }
-
-  // Set the appropriate stream
-  if (_type == PantomimeFormatMaildir)
-    {
-      aMailFile = [NSString stringWithFormat: @"%@:%@",
-		[NSString stringWithFormat: @"%lld.%d%d%lu.%@",
-						(long long)time(NULL), 
-						getpid(),
-						rand(),
-						(unsigned long)[_cacheManager count],
-						[[NSHost currentHost] name]],
-		((id)theFlags ? (id)[theFlags maildirString] : (id)@"2,")];
-      
-      aMailFilePath = [NSString stringWithFormat: 
-			@"%@/cur/%@", _path, aMailFile];
-
-#ifdef __MINGW32__      
-      aStream = fopen([aMailFilePath UTF8String], "bw+");
-#else
-      aStream = fopen([aMailFilePath UTF8String], "w+");
-#endif
-
-      if (!aStream)
-	{
-	  FOLDER_APPEND_FAILED;
-	}
-    }
-  else
-    {
-      aStream = [self stream];
-    }
-
-
-  // We keep the position where we were in the file
-  mark = ftell(aStream);
-
-  if (mark < 0)
-    {
-      FOLDER_APPEND_FAILED;
-    }
   
   //
   // If the message doesn't contain the "From ", we add it.
@@ -437,9 +384,9 @@
 	  [aMutableData replaceBytesInRange: aRange  withBytes: "\n From "];
 	  
 	  aRange = [aMutableData rangeOfCString: "\nFrom "
-				 options: 0
-				 range: NSMakeRange(aRange.location + aRange.length,
-						    [aMutableData length] - aRange.location - aRange.length) ];
+					options: 0
+					  range: NSMakeRange(aRange.location + aRange.length,
+							     [aMutableData length] - aRange.location - aRange.length) ];
 	}
   
       //
@@ -455,35 +402,43 @@
       //
       [aMutableData appendCString: "\n\n"];
     }
-  
-  // We go at the end of the file...
-  if (fseek(aStream, 0L, SEEK_END) < 0)
+
+    // We go at the end of the file...
+  if (fseek(theStream, 0L, SEEK_END) < 0)
     {
-      FOLDER_APPEND_FAILED;
+      NSLog(@"CWLocalfolder appendMessageFromRawSource: fseek() failed");
+      [aMutableData release];
+      return NO;
     }
   
   // We get the position of our message in the file. We need
   // to keep it in order to correctly seek back at the beginning
   // of the message to parse it.
-  if ((file_position = ftell(aStream)) < 0)
+  if ((file_position = ftell(theStream)) < 0)
     {
-      FOLDER_APPEND_FAILED;
+      NSLog(@"CWLocalfolder appendMessageFromRawSource: ftell() failed");
+      [aMutableData release];
+      return NO;
     }
       
   // We write the string to our local folder
-  if (fwrite([aMutableData bytes], 1, [aMutableData length], aStream) <= 0)
+  if (fwrite([aMutableData bytes], 1, [aMutableData length], theStream) <= 0)
     {
-      FOLDER_APPEND_FAILED;
+      NSLog(@"CWLocalfolder appendMessageFromRawSource: fwrite() failed");
+      [aMutableData release];
+      return NO;
     }
 
   // We parse the message using our code, which will also update
   // our cache if present
-  if (fseek(aStream, file_position, SEEK_SET) < 0)
+  if (fseek(theStream, file_position, SEEK_SET) < 0)
     {
-      FOLDER_APPEND_FAILED;
+      NSLog(@"CWLocalfolder appendMessageFromRawSource: fseek() failed");
+      [aMutableData release];
+      return NO;
     }
   
-  [self parse_mbox: aMailFile  stream: aStream  flags: theFlags  all: NO];
+  [self parse_mbox: mailFile  stream: theStream  flags: theFlags  all: NO];
   
   // We get back our message
   aMessage = [allMessages objectAtIndex: [allMessages count]-1];
@@ -494,28 +449,114 @@
       [aMessage setFlags: theFlags];
     }
 
-  // We finally reset our fp where the mark was set
-  if (_type != PantomimeFormatMaildir)
+   [aMutableData release];
+   return YES;
+}
+
+
+//
+// This method is used to append a message to this folder. The message
+// must be specified in raw source. The message is appended to the 
+// local file and is initialized after.
+//
+- (void) appendMessageFromRawSource: (NSData *) theData
+                              flags: (CWFlags *) theFlags
+{
+  NSString *aMailFile, *aMailFilePath;
+  NSDictionary *aDictionary;
+  FILE *aStream;
+  BOOL appendResult;
+  NSAutoreleasePool *arp;
+  long mark;
+
+  aMailFile = nil;
+  aStream = NULL;
+  mark = 0L;
+
+  aDictionary = (theFlags ? [NSDictionary dictionaryWithObjectsAndKeys: theData, @"NSData", self, @"Folder", theFlags, @"Flags", nil] :
+		 [NSDictionary dictionaryWithObjectsAndKeys: theData, @"NSData", self, @"Folder", nil]);
+
+  if (!theData || [theData length] == 0)
     {
-      if (fseek(aStream, mark, SEEK_SET) < 0)
+      NSLog(@"CWLocalfolder appendMessageFromRawSource: no data");
+      PERFORM_SELECTOR_3([[self store] delegate], @selector(folderAppendFailed:), PantomimeFolderAppendFailed, aDictionary);
+      return;
+    }
+
+  // Set the appropriate stream
+  if (_type == PantomimeFormatMaildir)
+    {
+      aMailFile = [NSString stringWithFormat: @"%@:%@",
+		[NSString stringWithFormat: @"%lld.%d%d%lu.%@",
+						(long long)time(NULL), 
+						getpid(),
+						rand(),
+						(unsigned long)[_cacheManager count],
+						[[NSHost currentHost] name]],
+		((id)theFlags ? (id)[theFlags maildirString] : (id)@"2,")];
+      
+      aMailFilePath = [NSString stringWithFormat: 
+			@"%@/cur/%@", _path, aMailFile];
+
+#ifdef __MINGW32__      
+      aStream = fopen([aMailFilePath UTF8String], "bw+");
+#else
+      aStream = fopen([aMailFilePath UTF8String], "w+");
+#endif
+      if (!aStream)
 	{
-	  FOLDER_APPEND_FAILED;
+          NSLog(@"CWLocalfolder appendMessageFromRawSource: fopen() failed");
+	  PERFORM_SELECTOR_3([[self store] delegate], @selector(folderAppendFailed:), PantomimeFolderAppendFailed, aDictionary);
+	  return;
 	}
     }
   else
     {
-      if (fclose(aStream) != 0)
-	{
-	  FOLDER_APPEND_FAILED;
-	}
+      // since we suppose it is already open, we do not close it afterwards
+      aStream = [self stream];
+      
+      // We keep the position where we were in the file
+      mark = ftell(aStream);
+      if (mark < 0)
+        {
+          NSLog(@"CWLocalfolder appendMessageFromRawSource: ftell() failed");
+          PERFORM_SELECTOR_3([[self store] delegate], @selector(folderAppendFailed:), PantomimeFolderAppendFailed, aDictionary);
+          return;
+        }      
     }
   
-  aDictionary = (theFlags ? [NSDictionary dictionaryWithObjectsAndKeys: theData, @"NSData", self, @"Folder", theFlags, @"Flags", nil] :
-		 [NSDictionary dictionaryWithObjectsAndKeys: theData, @"NSData", self, @"Folder", nil]);
-  PERFORM_SELECTOR_3([[self store] delegate], @selector(folderAppendCompleted:), PantomimeFolderAppendCompleted, aDictionary);
 
-  RELEASE(aMutableData);
-  RELEASE(pool);
+  arp = [[NSAutoreleasePool alloc] init];
+  appendResult = [self appendMessageFromRawSource:theData toStream:aStream withMailFile: aMailFile flags:theFlags];
+  [arp release];
+
+  if (appendResult)
+    {
+      PERFORM_SELECTOR_3([[self store] delegate], @selector(folderAppendCompleted:), PantomimeFolderAppendCompleted, aDictionary);
+    }
+  else
+    {
+      PERFORM_SELECTOR_3([[self store] delegate], @selector(folderAppendFailed:), PantomimeFolderAppendFailed, aDictionary);
+    }
+
+  // should we close it earlier before performing the selector?
+  // If opened the stream, we close it
+  if (_type == PantomimeFormatMaildir)
+    {
+      if (fclose(aStream) != 0)
+        {
+	  NSLog(@"CWLocalfolder appendMessageFromRawSource: fclose() failed");
+        }
+    }
+  // else we seek back
+  else
+    {
+      if (fseek(aStream, mark, SEEK_SET) < 0)
+	{
+          NSLog(@"CWLocalfolder appendMessageFromRawSource: resetting fp, fseek() failed");
+	}
+      
+    }
 }
 
 
@@ -531,7 +572,7 @@
   NSDictionary *userInfo;
   CWLocalMessage *aMessage;
 
-  int i, count;
+  NSUInteger i, count;
 
   aMutableArray = [NSMutableArray array];
 
@@ -714,7 +755,7 @@
       // The part content contains many part; we parse each part
       CWMIMEMultipart *aMimeMultipart;
       CWPart *aPart;
-      int i, count;
+      NSUInteger i, count;
       
       aMimeMultipart = (CWMIMEMultipart*)[thePart content];
       count = [aMimeMultipart count];
