@@ -1,9 +1,12 @@
 /*
 **  CWMIMEUtility.m
 **
-**  Copyright (c) 2001-2007
+**  Copyright (c) 2001-2007 Ludovic Marcotte
+**  Copyright (C) 2014-2018 Riccardo Mottola
 **
 **  Author: Ludovic Marcotte <ludovic@Sophos.ca>
+**          Riccardo Mottola <rm@gnu.org>
+**          Sebastian Reitenbach
 **
 **  This library is free software; you can redistribute it and/or
 **  modify it under the terms of the GNU Lesser General Public
@@ -15,29 +18,28 @@
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 **  Lesser General Public License for more details.
 **  
-**  You should have received a copy of the GNU Lesser General Public
-**  License along with this library; if not, write to the Free Software
-**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+** You should have received a copy of the GNU General Public License
+** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include <Pantomime/CWMIMEUtility.h>
+#import <Pantomime/CWMIMEUtility.h>
 
-#include <Pantomime/CWConstants.h>
-#include <Pantomime/CWMessage.h>
-#include <Pantomime/CWMIMEMultipart.h>
-#include <Pantomime/NSString+Extensions.h>
-#include <Pantomime/NSData+Extensions.h>
-#include <Pantomime/CWPart.h>
-#include <Pantomime/CWMD5.h>
-#include <Pantomime/CWUUFile.h>
+#import <Pantomime/CWConstants.h>
+#import <Pantomime/CWMessage.h>
+#import <Pantomime/CWMIMEMultipart.h>
+#import <Pantomime/NSString+Extensions.h>
+#import <Pantomime/NSData+Extensions.h>
+#import <Pantomime/CWPart.h>
+#import <Pantomime/CWMD5.h>
+#import <Pantomime/CWUUFile.h>
 
-#include <Foundation/NSAutoreleasePool.h>
-#include <Foundation/NSBundle.h>
-#include <Foundation/NSCharacterSet.h>
-#include <Foundation/NSFileManager.h>
-#include <Foundation/NSHost.h>
-#include <Foundation/NSScanner.h>
-#include <Foundation/NSValue.h>
+#import <Foundation/NSAutoreleasePool.h>
+#import <Foundation/NSBundle.h>
+#import <Foundation/NSCharacterSet.h>
+#import <Foundation/NSFileManager.h>
+#import <Foundation/NSHost.h>
+#import <Foundation/NSScanner.h>
+#import <Foundation/NSValue.h>
 
 #include <ctype.h>
 #include <unistd.h>
@@ -53,7 +55,7 @@ char *striphtml(char *s, int encoding);
 NSString *unique_id();
 
 static const char *hexDigit = "0123456789ABCDEF";
-static int seed_count = 1;
+static unsigned seed_count = 1;
 
 @implementation CWMIMEUtility
 
@@ -73,7 +75,7 @@ static int seed_count = 1;
   NSData *charset, *encoded_text;
   NSString *aString;
   
-  int i, length, start, i_charset, i_encoding, end; 
+  NSUInteger i, length, start, i_charset, i_encoding, end; 
   const char *bytes;
  
   BOOL ignore_span;
@@ -187,7 +189,9 @@ static int seed_count = 1;
 	}
       else
 	{
-	  charset = [theData subdataWithRange: NSMakeRange(start+2,i_charset-start-2)];
+          NSRange charsetRange;
+          charsetRange = NSMakeRange(start+2,i_charset-start-2);
+	  charset = [theData subdataWithRange: charsetRange];
 	}
       
       encoded_text = [theData subdataWithRange: NSMakeRange(i_encoding,end-i_encoding)];
@@ -199,7 +203,10 @@ static int seed_count = 1;
 	}
       else if (encoding == 'b' || encoding== 'B')
 	{
-	  aString = [NSString stringWithData: [encoded_text decodeBase64]
+          NSData *decodedData;
+        
+          decodedData = [encoded_text decodeBase64];
+	  aString = [NSString stringWithData: decodedData
 			      charset: charset];
 	}
       else
@@ -207,20 +214,18 @@ static int seed_count = 1;
 	  continue;
 	}
 
-      if (!aString)
+      if (aString)
 	{
-	  continue;
-	}
-      
-      [aMutableString appendString: aString];
+          [aMutableString appendString: aString];
+        }
       aString = nil;
       start = i;
       ignore_span = YES;
     }
   
+  // always append end of string, even if there was a space
   i = length;
-  
-  if (i != start && !ignore_span)
+  if (i > start)
     {
       aString = nil;
 
@@ -376,7 +381,7 @@ static int seed_count = 1;
 //
 #warning remove/combine or leave and document
 + (NSData *) encodeWordUsingQuotedPrintable: (NSString *) theWord
-			       prefixLength: (int) thePrefixLength
+			       prefixLength: (NSUInteger) thePrefixLength
 {  
   NSMutableString *aMutableString;
   NSMutableArray *aMutableArray;
@@ -384,7 +389,8 @@ static int seed_count = 1;
   NSMutableData *aMutableData;
   NSScanner *aScanner;
   
-  int i, count, previousLocation, currentLocation;
+  NSUInteger i, count;
+  NSUInteger previousLocation, currentLocation;
   BOOL mustUseEncoding;
 
   if (!theWord || [theWord length] == 0 )
@@ -414,7 +420,7 @@ static int seed_count = 1;
   
   while ([aScanner scanUpToCharactersFromSet: [NSCharacterSet whitespaceCharacterSet]  intoString: NULL])
     {
-      int length;
+      NSUInteger length;
       
       currentLocation = [aScanner scanLocation]; 
 	  
@@ -524,7 +530,7 @@ static int seed_count = 1;
   NSMutableData *aMutableData;
   NSArray *allParts;
   NSRange aRange;
-  int i, count;
+  NSUInteger i, count;
   
   aMimeMultipart = [[CWMIMEMultipart alloc] init];
 
@@ -763,7 +769,7 @@ char *striphtml(char *s, int encoding)
 	      if (aData)
 		{
 		  char *bytes;
-		  int len;
+		  NSUInteger len;
 		  
 		  bytes = (char *)[aData bytes];
 		  len = [aData length];
@@ -782,7 +788,7 @@ char *striphtml(char *s, int encoding)
         }
     }
 
-  *t++ = '\0';
+  *t = '\0';
 
   return text;
 }
@@ -794,7 +800,7 @@ char *striphtml(char *s, int encoding)
 //
 char ent(char **ref)
 {
-  int i;
+  unsigned i;
   char c = ' ', *s = *ref, *t = s;
   
   struct {
@@ -903,7 +909,7 @@ char ent(char **ref)
     { "yuml"  , (char)255 }
   };
   
-  while (isalpha(*s) || isdigit(*s) || *s == '#')
+  while (isalpha((int)(unsigned char)*s) || isdigit((int)(unsigned char)*s) || *s == '#')
     s++;
   
   for (i = 0; i < sizeof(refs)/sizeof(refs[0]); i++) {
@@ -944,7 +950,7 @@ NSString *unique_id()
   random_data[8] = '\0';
   
   aMutableData = [[NSMutableData alloc] init];
-  [aMutableData appendCFormat: @"%d.%d%s", pid, curtime, random_data];
+  [aMutableData appendCFormat: @"%d.%ld%s", pid, (long)curtime, random_data];
   aMD5 = [[CWMD5 alloc] initWithData: aMutableData];
   RELEASE(aMutableData);
   AUTORELEASE(aMD5);
