@@ -1,9 +1,13 @@
 /*
 **  CWParser.m
 **
-**  Copyright (c) 2001-2007
+**  Copyright (c) 2001-2007 Ludovic Marcotte
+**  Copyright (C) 2013-2018 The GNUstep Team
 **
 **  Author: Ludovic Marcotte <ludovic@Sophos.ca>
+**          Riccardo Mottola <rm@gnu.org>
+**          Sebastian Reitenbach
+**          German Arias
 **
 **  This library is free software; you can redistribute it and/or
 **  modify it under the terms of the GNU Lesser General Public
@@ -15,19 +19,19 @@
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 **  Lesser General Public License for more details.
 **  
-**  You should have received a copy of the GNU Lesser General Public
-**  License along with this library; if not, write to the Free Software
-**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+** You should have received a copy of the GNU General Public License
+** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
-#include <Pantomime/CWParser.h>
 
-#include <Pantomime/CWConstants.h>
-#include <Pantomime/CWFlags.h>
-#include <Pantomime/CWInternetAddress.h>
-#include <Pantomime/CWMessage.h>
-#include <Pantomime/CWMIMEUtility.h>
-#include <Pantomime/NSData+Extensions.h>
-#include <Pantomime/NSString+Extensions.h>
+#import <Pantomime/CWParser.h>
+
+#import <Pantomime/CWConstants.h>
+#import <Pantomime/CWFlags.h>
+#import <Pantomime/CWInternetAddress.h>
+#import <Pantomime/CWMessage.h>
+#import <Pantomime/CWMIMEUtility.h>
+#import <Pantomime/NSData+Extensions.h>
+#import <Pantomime/NSString+Extensions.h>
 
 #include <stdlib.h>
 #include <string.h>  // For NULL on OS X
@@ -35,10 +39,10 @@
 #include <stdio.h>
 //#include <Pantomime/elm_defs.h>
 
-#include <Foundation/NSBundle.h>
-#include <Foundation/NSTimeZone.h>
-#include <Foundation/NSString.h>
-#include <Foundation/NSURL.h>
+#import <Foundation/NSBundle.h>
+#import <Foundation/NSTimeZone.h>
+#import <Foundation/NSString.h>
+#import <Foundation/NSURL.h>
 
 //
 //
@@ -95,7 +99,7 @@ static struct _timezone {
 int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned char *word)
 {
   unsigned char *p;
-  int i;
+  unsigned i;
 
   for (p = buf+start, i = start; (isspace(*p) || *p == ','); ++p, ++i);
   
@@ -109,7 +113,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
   
   *word = '\0';
 
-  return p-buf-start;
+  return (p-buf)-start;
 }
 
 //
@@ -259,7 +263,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 {
   NSRange aRange;
   NSData *aData;
-  int x;
+  NSUInteger x;
 
   if ([theLine length] <= 14)
     {
@@ -268,10 +272,15 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
     }
 
   aData = [[theLine subdataFromIndex: 13] dataByTrimmingWhiteSpaces];
+  if (aData == nil || [aData length] == 0)
+    {
+      [thePart setContentType: @"text/plain"];
+      return;
+    }
 
   // We first skip the parameters, if we need to
   x = [aData indexOfCharacter: ';'];
-  if (x > 0)
+  if (x != NSNotFound && x > 0)
     {
       aData = [aData subdataToIndex: x];
     } 
@@ -279,7 +288,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
   // We see if there's a subtype specified for text, if none was specified, we append "/plain"
   x = [aData indexOfCharacter: '/'];
 
-  if (x < 0 && [aData hasCaseInsensitiveCPrefix: "text"])
+  if (x == NSNotFound && [aData hasCaseInsensitiveCPrefix: "text"])
     {
       [thePart setContentType: [[[aData asciiString] stringByAppendingString: @"/plain"] lowercaseString]];
     }
@@ -357,18 +366,20 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
   if ([theLine length] > 6)
     {
       NSData *aData;
-
-      int month, day, year, hours, mins, secs, tz, i, j, len, tot, s;
+      NSUInteger tot;
+      int month, day, year, hours, mins, secs, tz;
+      int len, s;
+      unsigned i, j;
       unsigned char *bytes, *word;
 
       aData = [theLine subdataFromIndex: 6];
       word = malloc(256);
-      
-      //NSLog(@"Have to parse |%@|", [aData asciiString]);
 
       bytes = (unsigned char*)[aData bytes];
       tot = [aData length];
-      i = len = 0;
+      i = 0;
+      s = 0;
+      tz = 0;
 
       // date-time       =       [ day-of-week "," ] date FWS time [CFWS]
       // day-of-week     =       ([FWS] day-name) / obs-day-of-week
@@ -415,11 +426,11 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 	  //NSLog(@"UNIX DATE");
 	  
 	  // We skip the first word, no need for it.
-	  i += len+1; len = next_word(bytes, i, tot, word); if (len <= 0) { free(word); return; }
+	  i += (unsigned)len+1; len = next_word(bytes, i, tot, word); if (len <= 0) { free(word); return; }
 	}
 
       
-      month = year = -1;
+      month = -1;
       
       // We got a RFC 822 date. The syntax is:
       // day month year hh:mm:ss zone
@@ -430,13 +441,13 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
       //printf("len = %d |%s| day = %d\n", len, word, day);
       
       // We get the month name and we convert it.
-      i += len+1; len = next_word(bytes, i, tot, word); if (len <= 0) { free(word); return; }
+      i += (unsigned)len+1; len = next_word(bytes, i, tot, word); if (len <= 0) { free(word); return; }
       
       for (j = 0; j < 12; j++)
 	{
 	  if (strncasecmp((const char*)word, month_name[j], 3) == 0)
 	    {
-	      month = j+1;
+	      month = (int)j+1;
 	    }
 	}
       
@@ -445,7 +456,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
       //printf("len = %d |%s| month = %d\n", len, word, month);
       
       // We get the year.
-      i += len+1; len = next_word(bytes, i, tot, word); if (len <= 0) { free(word); return; }
+      i += (unsigned)len+1; len = next_word(bytes, i, tot, word); if (len <= 0) { free(word); return; }
       year = atoi((const char*)word);
       
       if (year < 70) year += 2000;
@@ -454,12 +465,12 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
       //printf("len = %d |%s| year = %d\n", len, word, year);
       
       // We parse the time using the hh:mm:ss format.
-      i += len+1; len = next_word(bytes, i, tot, word); if (len <= 0) { free(word); return; }
+      i += (unsigned)len+1; len = next_word(bytes, i, tot, word); if (len <= 0) { free(word); return; }
       sscanf((const char*)word, "%d:%d:%d", &hours, &mins, &secs);
       //printf("len = %d |%s| %d:%d:%d\n", len, word, hours, mins, secs);
       
       // We parse the timezone.
-      i += len+1; len = next_word(bytes, i, tot, word); 
+      i += (unsigned)len+1; len = next_word(bytes, i, tot, word);
       
       if (len <= 0)
 	{
@@ -477,7 +488,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 	      p++;
 	    }
 	  
-	  len = strlen((const char*)p);
+	  len = (int)strlen((const char*)p);
 	  
 	  if (isdigit(*p))
 	    {
@@ -504,11 +515,11 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 	}
       
       [theMessage setReceivedDate: [NSCalendarDate dateWithYear: year
-						   month: month
-						   day: day
-						   hour: hours
-						   minute: mins
-						   second: secs
+						   month: (NSUInteger) month
+						   day: (NSUInteger)day
+						   hour: (NSUInteger) hours
+						   minute: (NSUInteger) mins
+						   second: (NSUInteger) secs
 						   timeZone: [NSTimeZone timeZoneForSecondsFromGMT: tz]]];
       free(word);
     }
@@ -525,7 +536,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
   CWInternetAddress *anInternetAddress;
   NSData *aData;
 
-  int i, len, s_len, x, y;
+  NSUInteger i, len, x;
   unsigned char *bytes;
   BOOL b;
 
@@ -554,6 +565,10 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 	case PantomimeResentToRecipient:
 	  len = 11;
 	  break;
+        default:
+          NSDebugLog(@"CWParser parseDestination: forType: unknown type!");
+          len = 0;
+          break;
 	}
 
       // We skip over emtpy headers.
@@ -575,26 +590,34 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 
       if (*bytes == ',' || i == len-1)
 	{
+          NSData *headerData;
+          NSRange headerRange;
+        
 	  if (b)
 	    {
 	      bytes++;
 	      continue;
 	    }
 
-	  y = i;
-
 	  // We strip the trailing comma for all but the last entries.
-	  s_len = y-x;
-	  if (i == len-1) s_len++;
-
+	  if (i == len-1)
+            headerRange = NSMakeRange(x, i-x+1);
+          else
+            headerRange = NSMakeRange(x, i-x);
+          
+          headerData = [aData subdataWithRange: headerRange];
+          headerData = [headerData dataByTrimmingWhiteSpaces];
 	  anInternetAddress = [[CWInternetAddress alloc]
-				initWithString: [CWMIMEUtility decodeHeader: [[aData subdataWithRange: NSMakeRange(x, s_len)] dataByTrimmingWhiteSpaces]
+				initWithString: [CWMIMEUtility decodeHeader: headerData
 							       charset: [theMessage defaultCharset]]];
  
-	  [anInternetAddress setType: theType];
-	  [theMessage addRecipient: anInternetAddress];
-	  RELEASE(anInternetAddress);
-	  x = y+1;
+	  if (anInternetAddress != nil)
+            {
+              [anInternetAddress setType: theType];
+	      [theMessage addRecipient: anInternetAddress];
+	      RELEASE(anInternetAddress);
+            }
+	  x = i+1;
 	}
       
       bytes++;
@@ -644,7 +667,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 		      quick: (BOOL) theBOOL
 {
   NSData *aData;
-  int x, y;
+  NSUInteger x, y;
   
   if (theBOOL)
     {
@@ -667,11 +690,11 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
   x = [aData indexOfCharacter: ';'];
   y = [aData indexOfCharacter: ' '];
 
-  if (x > 0 && x < y)
+  if (x != NSNotFound && x > 0 && x < y)
     {
       aData = [aData subdataToIndex: x];
     }
-  else if (y > 0)
+  else if (y != NSNotFound && y > 0)
     {
       aData = [aData subdataToIndex: y];
     }
@@ -743,7 +766,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
       NSArray *allReferences;
       NSData *aReference;
       NSString *aString;
-      int i, count;
+      NSUInteger i, count;
       
       allReferences = [aData componentsSeparatedByCString: " "];
       count = [allReferences count];
@@ -818,9 +841,11 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 	      anInternetAddress = [[CWInternetAddress alloc]
 				    initWithString: [CWMIMEUtility decodeHeader: [[aData subdataWithRange: NSMakeRange(x, s_len)] dataByTrimmingWhiteSpaces]
 								   charset: [theMessage defaultCharset]]];
-	      
-	      [aMutableArray addObject: anInternetAddress];
-	      RELEASE(anInternetAddress);
+	      if (anInternetAddress != nil)
+                {
+	          [aMutableArray addObject: anInternetAddress];
+	          RELEASE(anInternetAddress);
+                }
 	      x = y+1;
 	    }
 	  
@@ -1028,7 +1053,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
   // charset = us-ascii
   // It can be terminated by ';' or end of line.
   
-  // Look the the first occurrence of ';'.
+  // Look for the first occurrence of ';'.
   // That marks the end of this key value pair.
   // If we don't find one, we set it to the end of the line.
   r1 = [theLine rangeOfCString: ";"
@@ -1075,7 +1100,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 	  // int end;
 	  
 	  // We check if we have a charset, in case of a multiline value.
-	  if ([theLine characterAtIndex: NSMaxRange(theRange)+2] == '=')
+	  if ([theLine characterAtIndex: NSMaxRange(theRange)+2] == '*')
 	    {
 	      has_charset = YES;
 	    }
@@ -1096,21 +1121,29 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 
 	      if (r1.location != NSNotFound)
 		{
-		  value_start = NSMaxRange(r1)+2;
-		  
-		  if ([theLine characterAtIndex: value_start+1] == '*') value_start++;
+		  value_start = NSMaxRange(r1);
+
+		  // now we may need to skip *=, when data is encoded like
+		  // filename*1*=%56%67;
+		  // else it is like
+		  // filename*1="xxxxx";
+		  if ([theLine characterAtIndex: value_start] == '*')
+			value_start += 2;
+		  else
+			value_start	+= 1;
+
 
 		  r2 = [theLine rangeOfCString: ";"  options: 0  range: NSMakeRange(NSMaxRange(r1), len-NSMaxRange(r1))];
 		  
 		  if (r2.length > 0)
 		    {
-		      value_end = r2.location-1;
+		      value_end = r2.location;
 		    }
 		  else
 		    {
 		      value_end = len;
 		    }
-		  
+
 		  [aMutableData appendData: [[[theLine subdataWithRange: NSMakeRange(value_start,value_end-value_start)]
 					       dataFromSemicolonTerminatedData] dataFromQuotedData]];
 		}
@@ -1119,20 +1152,6 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 		  break;
 		}
 	    }
-	      
-	      // We now search again for ";"
-	      //r2 = [theLine rangeOfCString: ";"
-	  //		options: 0
-	  //		range: NSMakeRange(end, len-end)];
-	  // 
-	  //if (r2.length > 0)
-	  //  {
-	  //    value_end = r2.location-1;
-	  // }
-	  //else
-	  //  {
-	  //   value_end = len-1;
-	  //}
 	}
       else if ([theLine characterAtIndex: NSMaxRange(theRange)+1] == '=')
 	{
@@ -1147,6 +1166,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 	  BOOL lang = NO;
 	  NSData *aCharset;
 	  
+          aCharset = nil;
 	  r1 = [aMutableData rangeOfCString: "'"];
 
 	  if (r1.location != NSNotFound)
@@ -1162,7 +1182,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 		  NSLog(@"WE'VE GOT A LANGUAGE!");
 		}
 
-	      aCharset = [aMutableData subdataToIndex: r2.location-1];
+	      aCharset = [aMutableData subdataToIndex: r1.location];
 
 	      // We strip the charset and the language information from our data
 	      [aMutableData replaceBytesInRange: NSMakeRange(0, NSMaxRange(r2))
@@ -1173,7 +1193,7 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 	  if (theBOOL)
 	    {
 	      NSString *aString;
-	      
+
 	      aString = AUTORELEASE([[NSString alloc] initWithData: aMutableData  encoding: NSASCIIStringEncoding]);
 	      
 	      if (lang)
@@ -1187,6 +1207,14 @@ int next_word(unsigned char *buf, unsigned int start, unsigned int len, unsigned
 		}
 	    }
 	}
+      else
+        {
+          NSString *aString;
+        
+          aString = AUTORELEASE([[NSString alloc] initWithData: aMutableData  encoding: NSASCIIStringEncoding]);
+          
+          return aString;
+        }
     } 
   else
     {
