@@ -1,8 +1,8 @@
 /*
 **  NSData+Extensions.m
 **
-**  Copyright (c) 2001-2007
-**                2014
+**  Copyright (c) 2001-2007 Ludovic Marcotte
+**  Copyright (C) 2014-2018 Riccardo Mottola
 **
 **  Author: Ludovic Marcotte <ludovic@Sophos.ca>
 **          Riccardo Mottola <rm@gnu.org>
@@ -17,18 +17,17 @@
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 **  Lesser General Public License for more details.
 **  
-**  You should have received a copy of the GNU Lesser General Public
-**  License along with this library; if not, write to the Free Software
-**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+** You should have received a copy of the GNU General Public License
+** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include "Pantomime/NSData+Extensions.h"
+#import "Pantomime/NSData+Extensions.h"
 
-#include <Foundation/NSArray.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSString.h>
+#import <Foundation/NSArray.h>
+#import <Foundation/NSException.h>
+#import <Foundation/NSString.h>
 
-#include "Pantomime/CWConstants.h"
+#import "Pantomime/CWConstants.h"
 
 #include <stdlib.h>
 #include <string.h>
@@ -37,7 +36,7 @@
 // C functions and constants
 //
 int getValue(char c);
-void nb64ChunkFor3Characters(char *buf, const char *inBuf, int numChars);
+void nb64ChunkFor3Characters(char *buf, const char *inBuf, unsigned numChars);
 
 static const char basis_64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
 static const char *hexDigit = "0123456789ABCDEF";
@@ -57,7 +56,11 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 - (NSData *) decodeBase64
 {
-  int i, j, length, rawIndex, block, pad, data_len;
+  NSUInteger data_len;
+  NSUInteger i, j;
+  NSInteger length;
+  NSUInteger pad;
+  int rawIndex, block;
   const unsigned char *bytes;
   char *raw;
 
@@ -77,7 +80,7 @@ static const char *hexDigit = "0123456789ABCDEF";
   
   length = data_len * 6 / 8 - pad;
   
-  raw = (char *)malloc(length);
+  raw = (char *)malloc((size_t)length);
   rawIndex = 0;
 
   for (i = 0; i < data_len; i += 4)
@@ -116,30 +119,31 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 //
-- (NSData *) encodeBase64WithLineLength: (int) theLength
+- (NSData *) encodeBase64WithLineLength: (NSUInteger) theLength
 {
   const char *inBytes = [self bytes];
   const char *inBytesPtr = inBytes;
-  int inLength = [self length];
+  NSUInteger inLength = [self length];
+  NSUInteger i;
+  char *outBytes;
+  char *outBytesPtr;
 
-  char *outBytes = malloc(sizeof(char)*inLength*2);
-  char *outBytesPtr = outBytes;
-
-  int numWordsPerLine = theLength/4;
-  int wordCounter = 0;
+  NSUInteger numWordsPerLine = theLength/4;
+  NSUInteger wordCounter = 0;
 
+  outBytes = malloc(sizeof(char)*(size_t)inLength*2);
+  outBytesPtr = outBytes;
   // We memset 0 our buffer so with are sure to not have
   // any garbage in it.
-  memset(outBytes, 0, sizeof(char)*inLength*2);
+  memset(outBytes, 0, sizeof(char)*(size_t)inLength*2);
 
-  while (inLength > 0)
+  for (i = 0; i < inLength; i += 3)
     {
-      nb64ChunkFor3Characters(outBytesPtr, inBytesPtr, inLength);
+      nb64ChunkFor3Characters(outBytesPtr, inBytesPtr, (unsigned)(inLength-i));
       outBytesPtr += 4;
       inBytesPtr += 3;
-      inLength -= 3;
 
-      wordCounter ++;
+      wordCounter++;
 
       if (theLength && wordCounter == numWordsPerLine)
 	{
@@ -159,12 +163,12 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (NSData *) unfoldLines
 {
   NSMutableData *aMutableData;
-  int i, length;
+  NSUInteger i, length;
   
-  const unsigned char *bytes, *b;
+  const unsigned char *b;
   
   length = [self length];
-  b = bytes = [self bytes];
+  b = [self bytes];
   
   aMutableData = [[NSMutableData alloc] initWithCapacity: length];
   
@@ -211,6 +215,11 @@ static const char *hexDigit = "0123456789ABCDEF";
 	  b++,i++;
 	  continue;
 	}
+      else if (*b=='=' && i+1==len)
+        {
+          /* this is the last =, it is a soft-line-break at the end of the message */
+          continue;
+        }
       else if (*b=='=' && i+2<len)
 	{
 	  b++,i++;
@@ -228,6 +237,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 	    }
           else
             {
+              [result release];
               [[NSException exceptionWithName:@"Pantomime Exception" reason:@"Hex data contained invalid char" userInfo:nil] raise];
               return nil;
             }
@@ -248,6 +258,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 	    }
           else
             {
+              [result release];
               [[NSException exceptionWithName:@"Pantomime Exception" reason:@"Hex data contained invalid char" userInfo:nil] raise];
               return nil;
             }
@@ -272,12 +283,12 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 //
-- (NSData *) encodeQuotedPrintableWithLineLength: (int) theLength
+- (NSData *) encodeQuotedPrintableWithLineLength: (NSUInteger) theLength
 					inHeader: (BOOL) aBOOL
 {
   NSMutableData *aMutableData;
   const unsigned char *b;
-  int i, length, line;
+  NSUInteger i, length, line;
   char buf[4];
   
   aMutableData = [[NSMutableData alloc] initWithCapacity: [self length]];
@@ -355,12 +366,12 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (NSRange) rangeOfData: (NSData *) theData
 {
   const char *b, *bytes, *str;
-  int i, len, slen;
+  NSUInteger i, len, slen;
   
   bytes = [self bytes];
   len = [self length];
 
-  if (!theData)
+  if (theData == nil || [theData length] == 0)
     {
       return NSMakeRange(NSNotFound,0);
     }
@@ -400,7 +411,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 -(NSRange) rangeOfCString: (const char *) theCString
-		  options: (unsigned int) theOptions
+		  options: (NSUInteger) theOptions
 {
   return [self rangeOfCString: theCString 
 	       options: theOptions 
@@ -412,20 +423,31 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 -(NSRange) rangeOfCString: (const char *) theCString
-		  options: (unsigned int) theOptions
+		  options: (NSUInteger) theOptions
 		    range: (NSRange) theRange
 {
   const char *b, *bytes;
-  int i, len, slen;
+  NSUInteger i, len, slen;
   
   if (!theCString)
     {
       return NSMakeRange(NSNotFound,0);
     }
+  if (theRange.location == NSNotFound)
+    {
+      NSDebugLog(@"rangeOfCString: invalid range location");
+      return NSMakeRange(NSNotFound,0);
+    }
   
   bytes = [self bytes];
   len = [self length];
-  slen = strlen(theCString);
+  slen = (NSUInteger)strlen(theCString);
+  
+  if (len < slen)
+    {
+      NSDebugLog(@"rangeOfCString: string too small to compare %@ to %s", self, theCString);
+      return NSMakeRange(NSNotFound,0);
+    }
   
   b = bytes;
   
@@ -469,8 +491,11 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 //
-- (NSData *) subdataFromIndex: (int) theIndex
+- (NSData *) subdataFromIndex: (NSUInteger) theIndex
 {
+  if (theIndex > [self length])
+    return [NSData data];
+
   return [self subdataWithRange: NSMakeRange(theIndex, [self length] - theIndex)];
 }
 
@@ -478,7 +503,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 //
-- (NSData *) subdataToIndex: (int) theIndex
+- (NSData *) subdataToIndex: (NSUInteger) theIndex
 {
   return [self subdataWithRange: NSMakeRange(0, theIndex)];
 }
@@ -490,20 +515,26 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (NSData *) dataByTrimmingWhiteSpaces
 {
   const char *bytes;
-  int i, j, len;
+  NSUInteger i, len;
+  NSUInteger j;
   
   bytes = [self bytes];
+  if (bytes == NULL)
+    return [NSData data];
+  
   len = [self length];
+  if (len == 0)
+    return [NSData data];
 
   for (i = 0; i < len && (bytes[i] == ' ' || bytes[i] == '\t'); i++) ;
-  for (j = len-1; j >= 0 && (bytes[j] == ' ' || bytes[i] == '\t'); j--) ;
+
+  // We found only spaces, no need to search from the end
+  if (i == len)
+    return [NSData data];
   
-  if (j <= i)
-    {
-      return AUTORELEASE(RETAIN(self));
-    }
+  for (j = len; j > 0 && (bytes[j-1] == ' ' || bytes[j-1] == '\t'); j--) ;
       
-  return [self subdataWithRange: NSMakeRange(i, j-i+1)];
+  return [self subdataWithRange: NSMakeRange(i, j-i)];
 }
 
 
@@ -514,7 +545,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 {
   NSMutableData *aMutableData;
   const char *bytes;
-  int i, j, len;
+  NSUInteger i, j, len;
   char *dest;
   
   bytes = [self bytes];
@@ -545,7 +576,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (NSData *) dataFromQuotedData
 {
   const char *bytes;
-  int len;
+  NSUInteger len;
   
   bytes = [self bytes];
   len = [self length];
@@ -570,7 +601,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (NSData *) dataFromSemicolonTerminatedData
 {
   const char *bytes;
-  int len;
+  NSUInteger len;
   
   bytes = [self bytes];
   len = [self length];
@@ -591,10 +622,10 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 //
-- (int) indexOfCharacter: (char) theCharacter
+- (NSUInteger) indexOfCharacter: (unichar) theCharacter
 {
   const char *b;
-  int i, len;
+  NSUInteger i, len;
  
   b = [self bytes];
   len = [self length];
@@ -605,7 +636,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 	return i;
       }
   
-  return -1;
+  return NSNotFound;
 }
 
 
@@ -615,7 +646,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (BOOL) hasCPrefix: (const char *) theCString
 {
   const char *bytes;
-  int len, slen;
+  NSUInteger len, slen;
   
   if (!theCString)
     {
@@ -625,9 +656,9 @@ static const char *hexDigit = "0123456789ABCDEF";
   bytes = [self bytes];
   len = [self length];
 
-  slen = strlen(theCString);
+  slen = (NSUInteger)strlen(theCString);
   
-  if ( slen > len)
+  if ( slen == 0 ||  slen > len)
     {
       return NO;
     }
@@ -647,7 +678,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (BOOL) hasCSuffix: (const char *) theCString
 {
   const char *bytes;
-  int len, slen;
+  NSUInteger len, slen;
   
   if (!theCString) 
     {
@@ -657,9 +688,9 @@ static const char *hexDigit = "0123456789ABCDEF";
   bytes = [self bytes];
   len = [self length];
 
-  slen = strlen(theCString);
+  slen = (NSUInteger)strlen(theCString);
 
-  if (slen > len) 
+  if (slen == 0 || slen > len) 
     {
       return NO;
     }
@@ -679,7 +710,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (BOOL) hasCaseInsensitiveCPrefix: (const char *) theCString
 {
   const char *bytes;
-  int len, slen;
+  NSUInteger len, slen;
   
   if (!theCString) 
     {
@@ -688,9 +719,9 @@ static const char *hexDigit = "0123456789ABCDEF";
 
   bytes = [self bytes];
   len = [self length];
-  slen = strlen(theCString);
+  slen = (NSUInteger)strlen(theCString);
   
-  if ( slen > len)
+  if ( slen == 0 || slen > len)
     {
       return NO;
     }
@@ -710,7 +741,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (BOOL) hasCaseInsensitiveCSuffix: (const char *) theCString
 {
   const char *bytes;
-  int len, slen;
+  NSUInteger len, slen;
   
   if (!theCString)
     {
@@ -719,9 +750,9 @@ static const char *hexDigit = "0123456789ABCDEF";
   
   bytes = [self bytes];
   len = [self length];
-  slen = strlen(theCString);
+  slen = (NSUInteger)strlen(theCString);
 
-  if (slen > len) 
+  if (slen == 0 || slen > len) 
     {
       return NO;
     }  
@@ -740,7 +771,8 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 - (NSComparisonResult) caseInsensitiveCCompare: (const char *) theCString
 {
-  int slen, len, clen, i;
+  NSUInteger slen, len, clen;
+  int i;
   const char *bytes;
   
   // Is this ok?
@@ -751,7 +783,7 @@ static const char *hexDigit = "0123456789ABCDEF";
       
   bytes = [self bytes];
   len = [self length];
-  slen = strlen(theCString);
+  slen = (NSUInteger)strlen(theCString);
   
   if (slen > len)
     {
@@ -795,10 +827,13 @@ static const char *hexDigit = "0123456789ABCDEF";
 {
   NSMutableArray *aMutableArray;
   NSRange r1, r2;
-  int len;
+  NSUInteger len;
   
-  aMutableArray = [[NSMutableArray alloc] init];
   len = [self length];
+  if (len == 0)
+    return nil;
+
+  aMutableArray = [[NSMutableArray alloc] init];
   r1 = NSMakeRange(0,len);
   
   r2 = [self rangeOfCString: theCString
@@ -848,10 +883,11 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 //
-- (unichar) characterAtIndex: (int) theIndex
+- (unichar) characterAtIndex: (NSUInteger) theIndex
 {
   const char *bytes;
-  int i, len;
+  NSUInteger len;
+  unichar ch;
   
   len = [self length];
 
@@ -865,25 +901,22 @@ static const char *hexDigit = "0123456789ABCDEF";
     }
 
   bytes = [self bytes];
-
-  for (i = 0; i < theIndex; i++)
-    {
-      bytes++;
-    }
-
-  return (unichar)*bytes;
+  ch = (unichar)bytes[theIndex];
+  return ch;
 }
 
 
 //
 //
 //
-- (NSData *) unwrapWithLimit: (int) theQuoteLimit
+- (NSData *) unwrapWithLimit: (NSUInteger) theQuoteLimit
 {
   NSMutableData *aMutableData, *lines;
   NSData *aLine;
 
-  int i, len, quote_depth, line_quote_depth, line_start;
+  NSUInteger i, len;
+  NSInteger quote_depth, line_quote_depth;
+  NSUInteger line_start;
   BOOL is_flowed;
 
   len = [self length];
@@ -930,6 +963,24 @@ static const char *hexDigit = "0123456789ABCDEF";
       // We verify if the line ends with a soft break 
       is_flowed = [aLine length] > 0 && [aLine characterAtIndex: [aLine length]-1] == ' ';
 
+      // We verify if the line is not just empty
+      if (is_flowed)
+	{
+	  BOOL only_spaces;
+	  NSUInteger k;
+	  
+	  only_spaces = YES;
+	  k = 0;
+	  while (only_spaces && k < [aLine length]-1)
+	    {
+	      if ([aLine characterAtIndex:k] != ' ')
+		only_spaces = NO;
+	      k++;
+	    }
+
+	  is_flowed = !only_spaces;
+	}
+      
       // We must handle usenet signature as a special case
       if (is_flowed && [aLine caseInsensitiveCCompare: "-- "] == NSOrderedSame)
 	{
@@ -947,10 +998,10 @@ static const char *hexDigit = "0123456789ABCDEF";
 	  // The current line is flowed but has mis-matched quoting
 
 	  // We first append the previous paragraph to the buffer with the necessary quote characters
-	  if (quote_depth)
+	  if (quote_depth > 0)
 	    {
 	      [lines replaceBytesInRange: NSMakeRange(0, [lines length])
-		     withBytes: [[lines quoteWithLevel: quote_depth  wrappingLimit: theQuoteLimit] bytes]];
+		     withBytes: [[lines quoteWithLevel: (NSUInteger)quote_depth  wrappingLimit: theQuoteLimit] bytes]];
 	    }
 
 	  [aMutableData appendData: lines];
@@ -968,11 +1019,11 @@ static const char *hexDigit = "0123456789ABCDEF";
 	  [lines appendData: aLine];
 	  
 	  // We add the necessary quote characters in the paragraph
-	  if (quote_depth)
+	  if (quote_depth > 0)
 	    {
 	      NSData *d;
 	      
-	      d = [lines quoteWithLevel: quote_depth   wrappingLimit: theQuoteLimit];
+	      d = [lines quoteWithLevel: (NSUInteger)quote_depth   wrappingLimit: theQuoteLimit];
 	      [lines replaceBytesInRange: NSMakeRange(0, [lines length])
 		     withBytes: [d bytes]  length: [d length]];
 	    }
@@ -997,19 +1048,19 @@ static const char *hexDigit = "0123456789ABCDEF";
 	  // The line is fixed but has mis-matched quoting
 	  
 	  // We first append the previous paragraph (if any) to the buffer with the necessary quote characters
-	  if (quote_depth)
+	  if (quote_depth > 0)
 	    {
 	      [lines replaceBytesInRange: NSMakeRange(0, [lines length])
-		     withBytes: [[lines quoteWithLevel: quote_depth  wrappingLimit: theQuoteLimit] bytes]];
+		     withBytes: [[lines quoteWithLevel: (NSUInteger)quote_depth  wrappingLimit: theQuoteLimit] bytes]];
 	    }
 	  
 	  [aMutableData appendData: lines];
 	  [aMutableData appendCString: "\n"];
 
 	  // We append the fixed line to the buffer with the necessary quote characters
-	  if (line_quote_depth)
+	  if (line_quote_depth > 0)
 	    {
-	      aLine = [aLine quoteWithLevel: line_quote_depth  wrappingLimit: theQuoteLimit];
+	      aLine = [aLine quoteWithLevel: (NSUInteger)line_quote_depth  wrappingLimit: theQuoteLimit];
 	    }
 
 	  [aMutableData appendData: aLine];
@@ -1031,10 +1082,10 @@ static const char *hexDigit = "0123456789ABCDEF";
   // We must handle flowed lines that don't have a fixed line break at the end of the message
   if ([lines length])
     {
-      if (quote_depth)
+      if (quote_depth > 0)
 	{
 	  [lines replaceBytesInRange: NSMakeRange(0, [lines length])
-		 withBytes: [[lines quoteWithLevel: quote_depth  wrappingLimit: theQuoteLimit] bytes]];
+		 withBytes: [[lines quoteWithLevel: (NSUInteger)quote_depth  wrappingLimit: theQuoteLimit] bytes]];
 	}
 
       [aMutableData appendData: lines];
@@ -1050,13 +1101,14 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 //
-- (NSData *) wrapWithLimit: (int) theLimit
+- (NSData *) wrapWithLimit: (NSUInteger) theLimit
 {
   NSMutableData *aMutableData;
   NSData *aLine, *part;
   NSArray *lines;
-  int i, j, k, split;
-  int depth;
+  NSUInteger i, j, k;
+  NSUInteger split;
+  NSUInteger depth;
 
   // We first verify if the string is valid
   if ([self length] == 0)
@@ -1244,14 +1296,14 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 //
-- (NSData *) quoteWithLevel: (int) theLevel
-	      wrappingLimit: (int) theLimit
+- (NSData *) quoteWithLevel: (NSUInteger) theLevel
+	      wrappingLimit: (NSUInteger) theLimit
 {
   NSMutableData *aMutableData, *aQuotePrefix;
   NSData *aData, *aLine;
   NSArray *lines;
   BOOL isQuoted;
-  int i;
+  NSUInteger i;
 
   // We verify if the wrapping limit is smaller then the quote level
   if (theLevel > theLimit) 
@@ -1334,16 +1386,16 @@ static const char *hexDigit = "0123456789ABCDEF";
 //
 //
 - (void) insertCString: (const char *) theCString
-	       atIndex: (int) theIndex
+	       atIndex: (NSUInteger) theIndex
 {
-  int s_length, length;
+  NSUInteger s_length, length;
 
   if (!theCString)
     {
       return;
     }
   
-  s_length = strlen(theCString);
+  s_length = (NSUInteger)strlen(theCString);
 
   if (s_length == 0)
     {
@@ -1353,7 +1405,7 @@ static const char *hexDigit = "0123456789ABCDEF";
   length = [self length];
   
   // We insert at the beginning of the data
-  if (theIndex <= 0)
+  if (theIndex == 0)
     {
       NSMutableData *data;
       
@@ -1385,7 +1437,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 - (void) replaceCRLFWithLF
 {
   unsigned char *bytes, *bi, *bo;
-  int delta, i,length;
+  NSUInteger delta, i, length;
   
   bytes = [self mutableBytes];
   length = [self length];
@@ -1415,7 +1467,7 @@ static const char *hexDigit = "0123456789ABCDEF";
 {
   NSMutableData *aMutableData;
   unsigned char *bytes, *bi, *bo;
-  int delta, i, length;
+  NSUInteger delta, i, length;
   
   bi = bytes = [self mutableBytes];
   length = [self length];
@@ -1482,7 +1534,7 @@ int getValue(char c)
 //
 //
 //
-void nb64ChunkFor3Characters(char *buf, const char *inBuf, int theLength)
+void nb64ChunkFor3Characters(char *buf, const char *inBuf, unsigned theLength)
 {
   if (theLength >= 3)
     {
