@@ -1,9 +1,11 @@
 /*
 **  NSString+Extensions.m
 **
-**  Copyright (c) 2001-2007
+**  Copyright (c) 2001-2007 Ludovic Marcotte
+**  Copyright (C) 2014-2018 Riccardo Mottola
 **
 **  Author: Ludovic Marcotte <ludovic@Sophos.ca>
+**          Riccardo Mottola <rm@gnu.org>
 **
 **  This library is free software; you can redistribute it and/or
 **  modify it under the terms of the GNU Lesser General Public
@@ -15,20 +17,19 @@
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 **  Lesser General Public License for more details.
 **  
-**  You should have received a copy of the GNU Lesser General Public
-**  License along with this library; if not, write to the Free Software
-**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
+** You should have received a copy of the GNU General Public License
+** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
-#include <Pantomime/NSString+Extensions.h>
+#import <Pantomime/NSString+Extensions.h>
 
-#include <Pantomime/CWCharset.h>
-#include <Pantomime/CWConstants.h>
-#include <Pantomime/CWInternetAddress.h>
-#include <Pantomime/CWPart.h>
-#include <Pantomime/NSData+Extensions.h>
+#import <Pantomime/CWCharset.h>
+#import <Pantomime/CWConstants.h>
+#import <Pantomime/CWInternetAddress.h>
+#import <Pantomime/CWPart.h>
+#import <Pantomime/NSData+Extensions.h>
 
-#include <Foundation/NSBundle.h>
+#import <Foundation/NSBundle.h>
 
 //
 // We include the CoreFoundation headers under Mac OS X so we can support
@@ -37,14 +38,17 @@
 #ifdef MACOSX
 #include <CoreFoundation/CFString.h>
 #include <CoreFoundation/CFStringEncodingExt.h>
-#else
-#include <GNUstepBase/Additions.h>
 #endif
 
 #include <ctype.h>
 
 #ifdef HAVE_ICONV
 #include <iconv.h>
+#if defined (MACOSX) || defined (__NetBSD__) || defined (__FreeBSD__)
+#define iconv_const_qualifier const
+#else
+#define iconv_const_qualifier
+#endif
 #endif
 
 #define IS_PRINTABLE(c) (isascii(c) && isprint(c))
@@ -70,19 +74,41 @@
 //
 //
 //
-- (int) indexOfCharacter: (unichar) theCharacter
+- (NSUInteger) indexOfCharacter: (unichar) theCharacter
 {
-  return [self indexOfCharacter: theCharacter  fromIndex: 0];
+  return [self indexOfCharacter: theCharacter  fromIndex: 0u];
+}
+
+//
+//
+//
+- (NSUInteger) indexOfLastCharacter: (unichar) theCharacter
+{
+  NSUInteger len;
+  NSUInteger i;
+    
+  len = [self length];
+    
+  for (i = 1; i <= len; i++)
+    {
+      if ([self characterAtIndex: len-i] == theCharacter)
+        {
+          return len-i;
+        }
+    }
+    
+    return NSNotFound;
 }
 
 
 //
 //
 //
-- (int) indexOfCharacter: (unichar) theCharacter
-               fromIndex: (unsigned int) theIndex
+- (NSUInteger) indexOfCharacter: (unichar) theCharacter
+               fromIndex: (NSUInteger) theIndex
 {
-  int i, len;
+  NSUInteger len;
+  NSUInteger i;
   
   len = [self length];
   
@@ -94,7 +120,7 @@
 	}
     }
   
-  return -1;
+  return NSNotFound;
 }
 
 
@@ -131,7 +157,7 @@
 //
 - (NSString *) stringFromQuotedString
 {
-  int len;
+  NSUInteger len;
 
   len = [self length];
   
@@ -149,7 +175,7 @@
 //
 //
 //
-+ (NSString *) stringValueOfTransferEncoding: (int) theEncoding
++ (NSString *) stringValueOfTransferEncoding: (PantomimeEncoding) theEncoding
 {
   switch (theEncoding)
     {
@@ -174,7 +200,7 @@
 //
 //
 //
-+ (int) encodingForCharset: (NSData *) theCharset
++ (unsigned long) encodingForCharset: (NSData *) theCharset
 {
   return [self encodingForCharset: theCharset convertToNSStringEncoding: YES];
 }
@@ -182,11 +208,11 @@
 //
 // Convenience to be able to use CoreFoundation conversion instead of NSString
 //
-+ (int) encodingForCharset: (NSData *) theCharset
++ (unsigned long) encodingForCharset: (NSData *) theCharset
  convertToNSStringEncoding: (BOOL) shouldConvert
 {
   // We define some aliases for the string encoding.
-  static struct { NSString *name; int encoding; BOOL fromCoreFoundation; } encodings[] = {
+  static struct { NSString *name; unsigned long encoding; BOOL fromCoreFoundation; } encodings[] = {
     {@"ascii"         ,NSASCIIStringEncoding          ,NO},
     {@"us-ascii"      ,NSASCIIStringEncoding          ,NO},
     {@"default"       ,NSASCIIStringEncoding          ,NO},  // Ah... spammers.
@@ -218,6 +244,18 @@
     {@"gb2312"       ,NSGB2312StringEncoding                    ,NO},
     {@"utf-7"        ,NSUTF7StringEncoding                      ,NO},
     {@"unicode-1-1-utf-7", NSUTF7StringEncoding                 ,NO},  // To prever a bug (sort of) in MS Hotmail
+    {@"euc-kr"       ,NSKoreanEUCStringEncoding                 ,NO},
+#else // Core foundation support
+    {@"iso-8859-3"   ,kCFStringEncodingISOLatin3                ,YES},
+    {@"koi8-r"       ,kCFStringEncodingKOI8_R                   ,YES},
+    {@"gb2312"       ,kCFStringEncodingEUC_CN                   ,YES},  // tested instead of proper GB2312-80
+#if MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_4
+    {@"koi8-u"       ,kCFStringEncodingKOI8_U                   ,YES},
+#if MAC_OS_X_VERSION_MAX_ALLOWED > MAC_OS_X_VERSION_10_5
+    {@"utf-7"        ,kCFStringEncodingUTF7                     ,YES},
+#endif
+#endif
+    {@"euc-kr"       ,kCFStringEncodingEUC_KR                   ,YES},
 #endif
     {@"windows-1250" ,NSWindowsCP1250StringEncoding             ,NO},
     {@"windows-1251" ,NSWindowsCP1251StringEncoding             ,NO},
@@ -230,15 +268,14 @@
   };
   
   NSString *name;
-  int i;
+  unsigned i;
 
   name = [[NSString stringWithCString: [theCharset bytes] length: [theCharset length]] lowercaseString];
-  
   for (i = 0; i < sizeof(encodings)/sizeof(encodings[0]); i++)
     {
       if ([name isEqualToString: encodings[i].name])
         {
-          int enc = encodings[i].encoding;
+          unsigned long enc = encodings[i].encoding;
           // Under OS X, we use CoreFoundation if necessary to convert the encoding
           // to a NSString encoding.
 #ifdef MACOSX
@@ -246,7 +283,10 @@
             {
               if (shouldConvert)
 		{
-		  return CFStringConvertEncodingToNSStringEncoding(enc);
+                  NSStringEncoding nsEnc;
+                  
+                  nsEnc = CFStringConvertEncodingToNSStringEncoding (enc);
+                  return (unsigned long)nsEnc;
 		}
 	      else
 		{
@@ -255,13 +295,16 @@
 	    }
           else
             {
-              if (!shouldConvert)
+              if (shouldConvert)
 		{
-		  return CFStringConvertNSStringEncodingToEncoding(enc);
+                  return enc;
 		}
               else
 		{
-		  return enc;
+                  unsigned long enc2;
+                  
+                  enc2 = (unsigned long)CFStringConvertNSStringEncodingToEncoding(enc);
+                  return enc2;
 		}
 	    }
 #else
@@ -269,7 +312,7 @@
 #endif
         }
     }
-  
+
 #ifdef MACOSX
   // Last resort: try using CoreFoundation...
   CFStringEncoding enc;
@@ -288,13 +331,13 @@
     }
 #endif
   
-  return -1;
+  return NSNotFound;
 }
 
 //
 //
 //
-+ (int) encodingForPart: (CWPart *) thePart
++ (unsigned long) encodingForPart: (CWPart *) thePart
 {
   return [self encodingForPart: thePart convertToNSStringEncoding: YES];
 }
@@ -302,10 +345,10 @@
 //
 // Convenience to be able to use CoreFoundation conversion instead of NSString
 //
-+ (int)    encodingForPart: (CWPart *) thePart
++ (unsigned long)    encodingForPart: (CWPart *) thePart
  convertToNSStringEncoding: (BOOL) shouldConvert
 {
-  int encoding;
+  unsigned long encoding;
   
   // We get the encoding we are gonna use. We always favor the default encoding.
   if ([thePart defaultCharset])
@@ -323,7 +366,7 @@
       encoding = [NSString defaultCStringEncoding];
     }
   
-  if (encoding == -1 || encoding == NSASCIIStringEncoding)
+  if (encoding == NSNotFound || encoding == NSASCIIStringEncoding)
     {
       encoding = NSISOLatin1StringEncoding;
     }
@@ -338,7 +381,7 @@
 + (NSString *) stringWithData: (NSData *) theData
                       charset: (NSData *) theCharset
 {
-  int encoding;
+  unsigned long encoding;
   
   if (theData == nil)
     {
@@ -352,22 +395,23 @@
   encoding = [NSString encodingForCharset: theCharset];
 #endif
 
-  if (encoding == -1)
+  if (encoding == NSNotFound)
     {
 #ifdef HAVE_ICONV
       NSString *aString;
 
       const char *i_bytes, *from_code;
       char *o_bytes;
+      char *o_buff;
 
       size_t i_length, o_length, ret;
-      int total_length;
+      size_t total_length;
       iconv_t conv;
       
       // Instead of calling cString directly on theCharset, we first try
       // to obtain the ASCII string of the data object.
       from_code = [[theCharset asciiString] cString];
-      
+      NSLog(@"stringWithData - iconv fallback, charset: %s", from_code);
       if (!from_code)
 	{
 	  return nil;
@@ -375,7 +419,7 @@
       
       conv = iconv_open("UTF-8", from_code);
       
-      if ((int)conv < 0)
+      if (conv == (iconv_t)-1)
 	{
 	  // Let's assume we got US-ASCII here.
 	  return AUTORELEASE([[NSString alloc] initWithData: theData  encoding: NSASCIIStringEncoding]);
@@ -383,41 +427,63 @@
       
       i_bytes = [theData bytes];
       i_length = [theData length];
-      
-      total_length = o_length = sizeof(unichar)*i_length;
-      o_bytes = (char *)malloc(o_length);
-      
+
+      total_length = sizeof(unichar)*i_length;
+      o_length = total_length;
+      o_buff = (char *)malloc(o_length); // original pointer to the buffer
+      o_bytes = o_buff; // o_bytes is the working iconv pointer
+
       if (o_bytes == NULL) return nil;
 
       while (i_length > 0)
 	{
-	  ret = iconv(conv, (char **)&i_bytes, &i_length, &o_bytes, &o_length);
+	  ret = iconv(conv, (iconv_const_qualifier char **)&i_bytes, &i_length, &o_bytes, &o_length);
 	  
+	  // *i_bytes  Pointer to the byte just after the last character fetched.
+	  // *i_bytes  Number of remaining bytes in the source buffer.
+	  // *o_bytes  Pointer to the byte just after the last character stored.
+	  // *o_length Number of remainder bytes in the destination buffer.
 	  if (ret == (size_t)-1)
 	    {
-	      iconv_close(conv);
-	      
-	      total_length = total_length - o_length;
-	      o_bytes -= total_length;
-	      free(o_bytes);
-	      return nil;
+              // was our buffer too small?
+              if (errno == E2BIG)
+                {
+                  // realloc and try again
+		  total_length += sizeof(unichar);
+		  o_buff = realloc(o_buff, total_length);
+                  if (o_buff == NULL)
+                    {
+                      NSLog(@"stringWithData, realloc() failed while enlarging, returning nil");
+                      iconv_close(conv);
+                      return nil;
+                    }
+		  o_bytes = o_buff;
+		  o_length = total_length;
+                }
+              else
+                {
+                  // we could have had an invalid sequence EILSEQ or incomplete sequence EINVAL
+                  NSLog(@"stringWithData, invalid sequence, returning nil");
+	          iconv_close(conv);
+	          free(o_buff);
+	          return nil;
+                }
 	    }
 	}
-      
+
       total_length = total_length - o_length;
-      o_bytes -= total_length;
-      
-      // If we haven't used all our allocated buffer, we shrink it.
-      if (o_length > 0 && total_length > 0)
-	{
-	  realloc(o_bytes, total_length);
-	}
-      
-      aString = [[NSString alloc] initWithData: [NSData dataWithBytesNoCopy: o_bytes
-							length: total_length]
-				  encoding: NSUTF8StringEncoding];
+
+      // If we haven't used all our allocated buffer, we do not need shrink it.
+      // We just pass the reduced length to NSData and tell it not to free it
+      aString = [[NSString alloc] initWithData: [NSData dataWithBytesNoCopy: o_buff
+								     length: total_length
+							       freeWhenDone:NO]
+				      encoding: NSUTF8StringEncoding];
+
       iconv_close(conv);
       
+      // free the whole original buffer, NSData will be invalid, but aString remains fine
+      free(o_buff);
       return AUTORELEASE(aString);
 #else
       return nil;
@@ -426,28 +492,35 @@
 
 #ifdef MACOSX
   return AUTORELEASE((NSString *)CFStringCreateFromExternalRepresentation(NULL, (CFDataRef)theData, encoding));
-#else
+#else    
   return AUTORELEASE([[NSString alloc] initWithData: theData  encoding: encoding]);
-#endif
+#endif  
 }
 
 
 //
 //
-//
+// Return only charset which we can also later encoode (encodingForCharset:)
 #warning return Charset instead?
 - (NSString *) charset
 {
   NSMutableArray *aMutableArray;
   NSString *aString;
   CWCharset *aCharset;
-
-  unsigned int i, j;
+  NSUInteger i, j;
 
   aMutableArray = [[NSMutableArray alloc] initWithCapacity: 21];
 
   [aMutableArray addObject: [CWCharset charsetForName: @"iso-8859-1"]];
+  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1252"]];
   [aMutableArray addObject: [CWCharset charsetForName: @"iso-8859-2"]];
+  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1251"]];
+  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1253"]];
+  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1250"]];
+  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1254"]];
+#ifndef MACOSX
+  [aMutableArray addObject: [CWCharset charsetForName: @"koi8-r"]];
+  [aMutableArray addObject: [CWCharset charsetForName: @"koi8-u"]];
   [aMutableArray addObject: [CWCharset charsetForName: @"iso-8859-3"]];
   [aMutableArray addObject: [CWCharset charsetForName: @"iso-8859-4"]];
   [aMutableArray addObject: [CWCharset charsetForName: @"iso-8859-5"]];
@@ -460,41 +533,31 @@
   [aMutableArray addObject: [CWCharset charsetForName: @"iso-8859-13"]];
   [aMutableArray addObject: [CWCharset charsetForName: @"iso-8859-14"]];
   [aMutableArray addObject: [CWCharset charsetForName: @"iso-8859-15"]];
-  [aMutableArray addObject: [CWCharset charsetForName: @"koi8-r"]];
-  [aMutableArray addObject: [CWCharset charsetForName: @"koi8-u"]];
-  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1250"]];
-  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1251"]];
-  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1252"]];
-  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1253"]];
-  [aMutableArray addObject: [CWCharset charsetForName: @"windows-1254"]];
+#endif
 
-
-  for (i = 0; i < [self length]; i++)
+  // we look for possible charsets, either we find no possible ones
+  // or finish the string with a one or more charsets
+  i = 0;
+  while (i < [self length] && [aMutableArray count] > 0)
     {
-      for (j = 0; j < [aMutableArray count]; j++)
+      j = 0;
+      // For character we remove all charset it cannot be represented with
+      while (j < [aMutableArray count])
         {
-          if (![[aMutableArray objectAtIndex: j] characterIsInCharset: [self characterAtIndex: i]])
+          if ([[aMutableArray objectAtIndex: j] characterIsInCharset: [self characterAtIndex: i]])
+            {
+              j++;
+            }
+          else
             {
               // Character is not in the charset
               [aMutableArray removeObjectAtIndex: j];
-              j--;
             }
         }
-
-      // FIXME: can't break even if there is only one left. First we have to check
-      //        whether that encoding will actually work for the entire string. If it
-      //	doesn't we'll need to fall back to utf-8 (or something else that can encode
-      //        _everything_).
-      // 
-      // Intelligent string splitting would help, of course
-      //
-      if ([aMutableArray count] < 1)
-        {
-          // We have zero or one charset
-          break;
-        }
+      i++;
     }
 
+  NSDebugLog(@"We have %lu charsets: %@", (unsigned long) [aMutableArray count], aMutableArray);
   if ([aMutableArray count])
     {
       aCharset = [aMutableArray objectAtIndex: 0];
@@ -510,14 +573,13 @@
 	  aString = @"iso-2022-jp";
 	}
       else
-	{ 
+	{
 	  // We have no charset, we return a default charset
 	  aString = @"utf-8";
 	}
     }
 
   RELEASE(aMutableArray);
-  
   return aString;
 }
 
@@ -534,7 +596,7 @@
   const char *b;
   BOOL escaped;
   unichar ch;
-  int i, len;
+  NSUInteger i, len;
 
   //
   // We UTF-7 encode _only_ the non-ASCII parts.
@@ -553,7 +615,7 @@
 	}
       else
 	{
-	  int j;
+	  NSUInteger j;
 
 	  j = i+1;
 	  // We got a non-ASCII character, let's get the substring and encode it using UTF-7.
@@ -647,7 +709,7 @@
 
   BOOL escaped;
   unichar ch;
-  int i, len;
+  NSUInteger i, len;
 
   aMutableData = [[NSMutableData alloc] init];
   AUTORELEASE(aMutableData);
@@ -733,7 +795,7 @@
                                          withCharacter: (unichar) theReplacement
 {
   NSMutableString *aMutableString;
-  int len, i;
+  NSUInteger len, i;
   unichar c;
 
   if (!theTarget || !theReplacement || theTarget == theReplacement)
@@ -765,17 +827,18 @@
 //
 //
 //
-- (NSString *) stringByDeletingLastPathComponentWithSeparator: (unsigned char) theSeparator
+- (NSString *) stringByDeletingLastPathComponentWithSeparator: (unichar) theSeparator
 {
-  int i, c;
+  NSUInteger i;
+  NSUInteger c;
   
   c = [self length];
 
-  for (i = c-1; i >= 0; i--)
+  for (i = c; i > 0; i--)
     {
-      if ([self characterAtIndex: i] == theSeparator)
+      if ([self characterAtIndex: i-1] == theSeparator)
 	{
-	  return [self substringToIndex: i];
+	  return [self substringToIndex: i-1];
 	}
     }
 
@@ -785,7 +848,7 @@
 //
 // 
 //
-- (NSString *) stringByDeletingFirstPathSeparator: (unsigned char) theSeparator
+- (NSString *) stringByDeletingFirstPathSeparator: (unichar) theSeparator
 {
   if ([self length] && [self characterAtIndex: 0] == theSeparator)
     {
@@ -800,7 +863,7 @@
 //
 - (BOOL) is7bitSafe
 {
-  int i, len;
+  NSUInteger i, len;
   
   // We search for a non-ASCII character.
   len = [self length];
@@ -890,7 +953,7 @@
 {
   CWInternetAddress *anInternetAddress;
   NSMutableString *aMutableString;
-  int i, count;
+  NSUInteger i, count;
   
   aMutableString = [[NSMutableString alloc] init];
   count = [theRecipients count];
